import numpy as np
import pytest
from nabu.processing.fft_cuda import get_available_fft_implems
from nabu.testutils import get_data, generate_tests_scenarios, compare_shifted_images
from nabu.cuda.utils import get_cuda_context, __has_pycuda__
from nabu.opencl.utils import get_opencl_context, __has_pyopencl__
from nabu.thirdparty.algotom_convert_sino import extend_sinogram

__has_cufft__ = False
if __has_pycuda__:
    avail_fft = get_available_fft_implems()
    __has_cufft__ = len(avail_fft) > 0
__has_pycuda__ = __has_pycuda__ and __has_cufft__  # need both for using Cuda backprojector


if __has_pycuda__:
    from nabu.reconstruction.fbp import CudaBackprojector
    from nabu.reconstruction.hbp import HierarchicalBackprojector
if __has_pyopencl__:
    from nabu.reconstruction.fbp_opencl import OpenCLBackprojector


scenarios = generate_tests_scenarios({"backend": ["cuda", "opencl"]})


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    file_desc = get_data("sino_halftomo.npz")
    cls.sino = file_desc["sinogram"] * 1e4
    cls.rot_center = file_desc["rot_center"]
    cls.tol = 5e-3
    if __has_pycuda__:
        cls.cuda_ctx = get_cuda_context()
    if __has_pyopencl__:
        cls.opencl_ctx = get_opencl_context("all")


@pytest.mark.usefixtures("bootstrap")
@pytest.mark.parametrize("config", scenarios)
class TestHalftomo:
    def _get_backprojector(self, config, *bp_args, **bp_kwargs):
        if config["backend"] == "cuda":
            if not (__has_pycuda__):
                pytest.skip("Need pycuda + scikit-cuda or vkfft")
            Backprojector = CudaBackprojector
            ctx = self.cuda_ctx
        else:
            if not (__has_pyopencl__):
                pytest.skip("Need pyopencl")
            Backprojector = OpenCLBackprojector
            ctx = self.opencl_ctx
            if config.get("opencl_use_textures", True) is False:
                # patch "extra_options"
                extra_options = bp_kwargs.pop("extra_options", {})
                extra_options["use_textures"] = False
                bp_kwargs["extra_options"] = extra_options
        return Backprojector(*bp_args, **bp_kwargs, backend_options={"ctx": ctx})

    def test_halftomo_right_side(self, config, sino=None, rot_center=None):
        if sino is None:
            sino = self.sino
        if rot_center is None:
            rot_center = self.rot_center

        sino_extended, rot_center_ext = extend_sinogram(sino, rot_center, apply_log=False)
        sino_extended *= 2  # compat. with nabu normalization

        backprojector_extended = self._get_backprojector(
            config,
            sino_extended.shape,
            rot_center=rot_center_ext,
            halftomo=False,
            padding_mode="edges",
            angles=np.linspace(0, 2 * np.pi, sino.shape[0], True),
            extra_options={"centered_axis": True},
        )
        ref = backprojector_extended.fbp(sino_extended)

        backprojector = self._get_backprojector(
            config,
            sino.shape,
            rot_center=rot_center,
            halftomo=True,
            padding_mode="edges",
            extra_options={"centered_axis": True},
        )
        res = backprojector.fbp(sino)

        # The approach in algotom (used as reference) slightly differers:
        #   - altogom extends the sinogram with padding, so that it's ready-to-use for FBP
        #   - nabu filters the sinogram first, and then does the "half-tomo preparation".
        #     Filtering the sinogram first is better to avoid artefacts due to sharp transition in the borders
        metric, upper_bound = compare_shifted_images(res, ref, return_upper_bound=True)
        assert metric < 5, "Something wrong for halftomo with backend %s" % (config["backend"])

    def test_halftomo_left_side(self, config):
        sino = np.ascontiguousarray(self.sino[:, ::-1])
        rot_center = sino.shape[-1] - 1 - self.rot_center
        return self.test_halftomo_right_side(config, sino=sino, rot_center=rot_center)

    def test_halftomo_cor_outside_fov(self, config):
        sino = np.ascontiguousarray(self.sino[:, : self.sino.shape[-1] // 2])
        backprojector = self._get_backprojector(config, sino.shape, rot_center=self.rot_center, halftomo=True)
        res = backprojector.fbp(sino)
        # Just check that it runs, but no reference results. Who does this anyway ?!

    @pytest.mark.skipif(not (__has_pycuda__), reason="Need pycuda")
    def test_hbp_halftomo(self, config):
        if config["backend"] == "opencl":
            pytest.skip("No HBP available in OpenCL")
        B = HierarchicalBackprojector(self.sino.shape, halftomo=True, rot_center=self.rot_center, padding_mode="edge")
        res = B.fbp(self.sino)

        sino_extended, rot_center_ext = extend_sinogram(self.sino, self.rot_center, apply_log=False)
        sino_extended *= 2  # compat. with nabu normalization
        B_extended = HierarchicalBackprojector(
            sino_extended.shape,
            rot_center=rot_center_ext,
            padding_mode="edge",
            angles=np.linspace(0, 2 * np.pi, self.sino.shape[0], True),
        )
        res_e = B_extended.fbp(sino_extended)

        # see notes in test_halftomo_right_side()
        metric, upper_bound = compare_shifted_images(res, res_e, return_upper_bound=True)
        assert metric < 5, "Something wrong for halftomo with HBP"



import numpy as np
import pytest
from nabu.testutils import get_data
from nabu.cuda.utils import __has_pycuda__

if __has_pycuda__:
    import pycuda.gpuarray as garray

    # from pycuda.cumath import fabs
    from pycuda.elementwise import ElementwiseKernel
    from nabu.cuda.utils import get_cuda_context
    from nabu.reconstruction.projection import Projector
    from nabu.reconstruction.fbp import Backprojector
try:
    import astra

    __has_astra__ = True
except ImportError:
    __has_astra__ = False


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    cls.image = get_data("brain_phantom.npz")["data"]
    cls.sino_ref = get_data("mri_sino500.npz")["data"]
    cls.n_angles, cls.dwidth = cls.sino_ref.shape
    cls.rtol = 1e-3
    if __has_pycuda__:
        cls.ctx = get_cuda_context()


@pytest.mark.skipif(not (__has_pycuda__), reason="Need pycuda for this test")
@pytest.mark.usefixtures("bootstrap")
class TestProjection:
    def check_result(self, img1, img2, err_msg):
        max_diff = np.max(np.abs(img1 - img2))
        assert max_diff / img1.max() < self.rtol, err_msg + " : max diff = %.3e" % max_diff

    def test_proj_simple(self):
        P = Projector(self.image.shape, self.n_angles)
        res = P(self.image)
        self.check_result(res, self.sino_ref, "Something wrong with simple projection")

    def test_input_output_kinds(self):
        P = Projector(self.image.shape, self.n_angles)

        # input on GPU, output on CPU
        d_img = garray.to_gpu(self.image)
        res = P(d_img)
        self.check_result(res, self.sino_ref, "Something wrong: input GPU, output CPU")

        # input on CPU, output on GPU
        out = garray.zeros(P.sino_shape, "f")
        res = P(self.image, output=out)
        self.check_result(out.get(), self.sino_ref, "Something wrong: input CPU, output GPU")

        # input and output on GPU
        out.fill(0)
        P(d_img, output=out)
        self.check_result(out.get(), self.sino_ref, "Something wrong: input GPU, output GPU")

    def test_odd_size(self):
        image = self.image[:511, :]
        P = Projector(image.shape, self.n_angles - 1)
        res = P(image)

    @pytest.mark.skipif(not (__has_astra__), reason="Need astra-toolbox for this test")
    def test_against_astra(self):
        def proj_astra(img, angles, rot_center=None):
            vol_geom = astra.create_vol_geom(img.shape)
            if np.isscalar(angles):
                angles = np.linspace(0, np.pi, angles, False)
            proj_geom = astra.create_proj_geom("parallel", 1.0, img.shape[-1], angles)
            if rot_center is not None:
                cor_shift = (img.shape[-1] - 1) / 2.0 - rot_center
                proj_geom = astra.geom_postalignment(proj_geom, cor_shift)

            projector_id = astra.create_projector("cuda", proj_geom, vol_geom)
            sinogram_id, sinogram = astra.create_sino(img, projector_id)

            astra.data2d.delete(sinogram_id)
            astra.projector.delete(projector_id)
            return sinogram

        # Center of rotation to test
        cors = [None, 255.5, 256, 260, 270.2, 300, 150]

        for cor in cors:
            res_astra = proj_astra(self.image, 500, rot_center=cor)
            res_nabu = Projector(self.image.shape, 500, rot_center=cor).projection(self.image)
            self.check_result(res_nabu, res_astra, "Projection with CoR = %s" % str(cor))

    @pytest.mark.skipif(not (__has_astra__), reason="Need astra-toolbox for this test")
    def test_em_reconstruction(self):
        """
        Test iterative reconstruction: Maximum Likelyhood Expectation Maximization (MLEM)
        """

        subsampling = 5
        sino = self.sino_ref[::subsampling, :]

        P = Projector(self.image.shape, sino.shape[0])
        B = Backprojector(sino.shape, padding_mode="edge", extra_options={"centered_axis": True})

        d_sino = garray.to_gpu(np.ascontiguousarray(sino))

        def EM(sino, P, B, n_it, eps=1e-6):
            ones = np.ones(sino.shape, "f")
            oinv = garray.to_gpu((1.0 / B.backproj(ones)).astype("f"))
            x = garray.ones_like(oinv)
            y = garray.zeros_like(x)
            proj = garray.zeros_like(sino)
            proj_inv = sino.copy()

            update_projection = ElementwiseKernel(
                "float* proj_inv, float* proj, float* proj_data, float eps",
                "proj_inv[i] = proj_data[i] / ((fabsf(proj[i]) > eps) ? (proj[i]) : (1.0f))",
                "update_projection",
            )
            for k in range(n_it):
                # proj = P(x)
                P.projection(x, output=proj)

                update_projection(proj_inv, proj, sino, eps)

                # x *= B(proj_inv) * oinv
                B.backproj(proj_inv, output=y)
                x *= y
                x *= oinv
            return x

        rec = EM(d_sino, P, B, 50)

        def EM_astra(sino, rec_shape, n_it):
            vol_geom = astra.create_vol_geom(rec_shape)
            proj_geom = astra.create_proj_geom(
                "parallel", 1.0, sino.shape[-1], np.linspace(0, np.pi, sino.shape[0], False)
            )

            rec_id = astra.data2d.create("-vol", vol_geom)
            sinogram_id = astra.data2d.create("-sino", proj_geom)
            astra.data2d.store(sinogram_id, sino)
            astra.data2d.store(rec_id, np.ones(rec_shape, "f"))  # !

            cfg = astra.astra_dict("EM_CUDA")
            cfg["ReconstructionDataId"] = rec_id
            cfg["ProjectionDataId"] = sinogram_id

            alg_id = astra.algorithm.create(cfg)
            astra.algorithm.run(alg_id, n_it)

            rec = astra.data2d.get(rec_id)

            astra.algorithm.delete(alg_id)
            astra.data2d.delete(rec_id)
            astra.data2d.delete(sinogram_id)

            return rec

        ref = EM_astra(sino, self.image.shape, 50)

        err_max = np.max(np.abs(rec.get() - ref))
        assert err_max < 0.2, "Discrepancy between EM and EM_astra"



import numpy as np
from .processing_base import ProcessingBase


class MulAdd:
    processing_cls = ProcessingBase

    def __init__(self, **backend_options):
        self.processing = self.processing_cls(**(backend_options or {}))
        self._init_finalize()

    def _init_finalize(self):
        pass

    def mul_add(self, dst, other, fac_dst, fac_other, dst_region=None, other_region=None):
        if dst_region is None:
            dst_slice_y = dst_slice_x = slice(None, None)
        else:
            dst_slice_y, dst_slice_x = dst_region
        if other_region is None:
            other_slice_y = other_slice_x = slice(None, None)
        else:
            other_slice_y, other_slice_x = other_region

        dst[dst_slice_y, dst_slice_x] = (
            fac_dst * dst[dst_slice_y, dst_slice_x] + fac_other * other[other_slice_y, other_slice_x]
        )

    __call__ = mul_add



import numpy as np
from ..utils import deprecated


def binning(img, bin_factor, out_dtype=np.float32):
    """
    Bin an image by a factor of "bin_factor".

    Parameters
    ----------
    bin_factor: tuple of int
        Binning factor in each axis.
    out_dtype: dtype, optional
        Output data type. Default is float32.

    Notes
    -----
    If the image original size is not a multiple of the binning factor,
    the last items (in the considered axis) will be dropped.
    The resulting shape is (img.shape[0] // bin_factor[0], img.shape[1] // bin_factor[1])
    """
    s = img.shape
    n0, n1 = bin_factor
    shp = (s[0] - (s[0] % n0), s[1] - (s[1] % n1))
    sub_img = img[: shp[0], : shp[1]]
    out_shp = (shp[0] // n0, shp[1] // n1)
    res = np.zeros(out_shp, dtype=out_dtype)
    for i in range(n0):
        for j in range(n1):
            res[:] += sub_img[i::n0, j::n1]
    res /= n0 * n1
    return res


def binning_n_alt(img, bin_factor, out_dtype=np.float32):
    """
    Alternate, "clever" but slower implementation
    """
    n0, n1 = bin_factor
    new_shape = tuple(s - (s % n) for s, n in zip(img.shape, bin_factor))
    sub_img = img[: new_shape[0], : new_shape[1]]
    img_view_4d = sub_img.reshape((new_shape[0] // n0, n0, new_shape[1] // n1, n1))
    return img_view_4d.astype(out_dtype).mean(axis=1).mean(axis=-1)


#
# COMPAT.
#


@deprecated("Please use binning()", do_print=True)
def binning2(img, out_dtype=np.float32):
    return binning(img, (2, 2), out_dtype=out_dtype)


@deprecated("Please use binning()", do_print=True)
def binning2_horiz(img, out_dtype=np.float32):
    return binning(img, (1, 2), out_dtype=out_dtype)


@deprecated("Please use binning()", do_print=True)
def binning2_vertic(img, out_dtype=np.float32):
    return binning(img, (2, 1), out_dtype=out_dtype)


@deprecated("Please use binning()", do_print=True)
def binning3(img, out_dtype=np.float32):
    return binning(img, (3, 3), out_dtype=out_dtype)


@deprecated("Please use binning()", do_print=True)
def binning3_horiz(img, out_dtype=np.float32):
    return binning(img, (1, 3), out_dtype=out_dtype)


@deprecated("Please use binning()", do_print=True)
def binning3_vertic(img, out_dtype=np.float32):
    return binning(img, (3, 1), out_dtype=out_dtype)


@deprecated("Please use binning()", do_print=True)
def get_binning_function(binning_factor):
    """
    Determine the binning function to use.
    """
    binning_functions = {
        (2, 2): binning2,
        (2, 1): binning2_vertic,
        (1, 2): binning2_horiz,
        (3, 3): binning3,
        (3, 1): binning3_vertic,
        (1, 3): binning3_horiz,
        (2, 3): None,  # was a limitation
        (3, 2): None,  # was a limitation
    }
    if binning_factor not in binning_functions:
        raise ValueError("Could not get a function for binning factor %s" % binning_factor)
    return binning_functions[binning_factor]



import pycuda.gpuarray as garray
from pycuda.compiler import SourceModule
from ..processing.kernel_base import KernelBase
from ..utils import catch_warnings  # TODO use warnings.catch_warnings once python < 3.11 is dropped


class CudaKernel(KernelBase):
    """
    Helper class that wraps CUDA kernel through pycuda SourceModule.

    Parameters
    -----------
    kernel_name: str
        Name of the CUDA kernel.
    filename: str, optional
        Path to the file name containing kernels definitions
    src: str, optional
        Source code of kernels definitions
    signature: str, optional
        Signature of kernel function. If provided, pycuda will not guess the types
        of kernel arguments, making the calls slightly faster.
        For example, a function acting on two pointers, an integer and a float32
        has the signature "PPif".
    texrefs: list, optional
        List of texture references, if any
    automation_params: dict, optional
        Automation parameters, see below
    sourcemodule_kwargs: optional
        Extra arguments to provide to pycuda.compiler.SourceModule(),

    """

    def __init__(
        self,
        kernel_name,
        filename=None,
        src=None,
        signature=None,
        texrefs=None,
        automation_params=None,
        silent_compilation_warnings=False,
        **sourcemodule_kwargs,
    ):
        super().__init__(
            kernel_name,
            filename=filename,
            src=src,
            automation_params=automation_params,
            silent_compilation_warnings=silent_compilation_warnings,
        )
        self.compile_kernel_source(kernel_name, sourcemodule_kwargs)
        self.prepare(signature, texrefs)

    def compile_kernel_source(self, kernel_name, sourcemodule_kwargs):
        self.sourcemodule_kwargs = sourcemodule_kwargs
        self.kernel_name = kernel_name
        with catch_warnings(action=("ignore" if self.silent_compilation_warnings else None)):  # pylint: disable=E1123
            self.module = SourceModule(self.src, **self.sourcemodule_kwargs)
        self.func = self.module.get_function(kernel_name)

    def prepare(self, kernel_signature, texrefs):
        self.prepared = False
        self.kernel_signature = kernel_signature
        self.texrefs = texrefs or []
        if kernel_signature is not None:
            self.func.prepare(self.kernel_signature, texrefs=self.texrefs)
            self.prepared = True

    def follow_device_arr(self, args):
        args = list(args)
        # Replace GPUArray with GPUArray.gpudata
        for i, arg in enumerate(args):
            if isinstance(arg, garray.GPUArray):
                args[i] = arg.gpudata
        return tuple(args)

    def get_last_kernel_time(self):
        """
        Return the execution time (in seconds) of the last called kernel.
        The last called kernel should have been called with time_kernel=True.
        """
        if self.last_kernel_time is not None:
            return self.last_kernel_time()
        else:
            return None

    def call(self, *args, **kwargs):
        grid, block, args, kwargs = self._prepare_call(*args, **kwargs)

        if self.prepared:
            func_call = self.func.prepared_call
            if "time_kernel" in kwargs:
                func_call = self.func.prepared_timed_call
                kwargs.pop("time_kernel")
            if "block" in kwargs:
                kwargs.pop("block")
            if "grid" in kwargs:
                kwargs.pop("grid")
            t = func_call(grid, block, *args, **kwargs)
        else:
            kwargs["block"] = block
            kwargs["grid"] = grid
            t = self.func(*args, **kwargs)
        self.last_kernel_time = t  # list ?

    __call__ = call



import numpy as np
from ..utils import get_cuda_srcfile, updiv
from ..cuda.processing import CudaProcessing
from ..cuda.utils import __has_pycuda__
from .padding_base import PaddingBase


class CudaPadding(PaddingBase):
    """
    A class for performing padding on GPU using Cuda
    """

    backend = "cuda"

    # TODO docstring from base class
    def __init__(self, shape, pad_width, mode="constant", cuda_options=None, **kwargs):
        super().__init__(shape, pad_width, mode=mode, **kwargs)
        self.cuda_processing = self.processing = CudaProcessing(**(cuda_options or {}))
        self._init_cuda_coordinate_transform()

    def _init_cuda_coordinate_transform(self):
        if self.mode == "constant":
            self.d_padded_array_constant = self.processing.to_device(
                "d_padded_array_constant", self.padded_array_constant
            )
            return
        self._coords_transform_kernel = self.processing.kernel(
            "coordinate_transform",
            filename=get_cuda_srcfile("padding.cu"),
            signature="PPPPiii",
        )
        self._coords_transform_block = (32, 32, 1)
        self._coords_transform_grid = [
            updiv(a, b) for a, b in zip(self.padded_shape[::-1], self._coords_transform_block)
        ]
        self.d_coords_rows = self.processing.to_device("d_coords_rows", self.coords_rows)
        self.d_coords_cols = self.processing.to_device("d_coords_cols", self.coords_cols)

    def _pad_constant(self, image, output):
        pad_y, pad_x = self.pad_width
        self.d_padded_array_constant[pad_y[0] : pad_y[0] + self.shape[0], pad_x[0] : pad_x[0] + self.shape[1]] = image[
            :
        ]
        output[:] = self.d_padded_array_constant[:]
        return output

    def pad(self, image, output=None):
        """
        Pad an array.

        Parameters
        ----------
        image: pycuda.gpuarray.GPUArray
            Image to pad
        output: pycuda.gpuarray.GPUArray, optional
            Output image. If provided, must be in the expected shape.
        """
        if output is None:
            output = self.processing.allocate_array("d_output", self.padded_shape)
        if self.mode == "constant":
            return self._pad_constant(image, output)
        self._coords_transform_kernel(
            image,
            output,
            self.d_coords_cols,
            self.d_coords_rows,
            np.int32(self.shape[1]),
            np.int32(self.padded_shape[1]),
            np.int32(self.padded_shape[0]),
            grid=self._coords_transform_grid,
            block=self._coords_transform_block,
        )
        return output

    __call__ = pad



from os import linesep
from configparser import ConfigParser
from ..utils import check_supported, deprecated

#
# option "type":
#  - required: always visible, user must provide a valid value
#  - optional: visible, but might be left blank
#  - advanced: optional and not visible by default
#  - unsupported: hidden (not implemented yet)
_options_levels = {
    "required": 0,
    "optional": 1,
    "advanced": 2,
    "unsupported": 10,
}


class NabuConfigParser:
    @deprecated(
        "The class 'NabuConfigParser' is deprecated and will be removed in a future version. Please use parse_nabu_config_file instead",
        do_print=True,
    )
    def __init__(self, fname):
        """
        Nabu configuration file parser.

        Parameters
        ----------
        fname: str
            File name of the configuration file
        """
        parser = ConfigParser(inline_comment_prefixes=("#",))  # allow in-line comments
        with open(fname) as fid:
            file_content = fid.read()
        parser.read_string(file_content)
        self.parser = parser
        self.get_dict()
        self.file_content = file_content.split(linesep)

    def get_dict(self):
        # Is there an officially supported way to do this ?
        self.conf_dict = self.parser._sections
        return self.conf_dict

    def __str__(self):
        return self.conf_dict.__str__()

    def __repr__(self):
        return self.conf_dict.__repr__()

    def __getitem__(self, key):
        return self.conf_dict[key]


def parse_nabu_config_file(fname, allow_no_value=False):
    """
    Parse a configuration file and returns a dictionary.

    Parameters
    ----------
    fname: str
        File name of the configuration file

    Returns
    -------
    conf_dict: dict
        Dictionary with the configuration
    """
    parser = ConfigParser(
        inline_comment_prefixes=("#",),  # allow in-line comments
        allow_no_value=allow_no_value,
    )
    with open(fname) as fid:
        file_content = fid.read()
    parser.read_string(file_content)
    conf_dict = parser._sections  # Is there an officially supported way to do this ?
    return conf_dict


def generate_nabu_configfile(
    fname,
    default_config,
    config=None,
    sections=None,
    sections_comments=None,
    comments=True,
    options_level=None,
    prefilled_values=None,
):
    """
    Generate a nabu configuration file.

    Parameters
    -----------
    fname: str
        Output file path.
    config: dict
        Configuration to save. If section and / or key missing will store the
        default value
    sections: list of str, optional
        Sections which should be included in the configuration file
    comments: bool, optional
        Whether to include comments in the configuration file
    options_level: str, optional
        Which "level" of options to embed in the file. Can be "required", "optional", "advanced".
        Default is "optional".
    """
    if options_level is None:
        options_level = "optional"
    if prefilled_values is None:
        prefilled_values = {}
    check_supported(options_level, list(_options_levels.keys()), "options_level")
    options_level = _options_levels[options_level]
    if config is None:
        config = {}
    if sections is None:
        sections = default_config.keys()

    def dump_help(fid, help_sequence):
        for help_line in help_sequence.split(linesep):
            content = "# %s" % (help_line) if help_line.strip() != "" else ""
            content = content + linesep
            fid.write(content)

    with open(fname, "w") as fid:
        for section, section_content in default_config.items():
            if section not in sections:
                continue
            if section != "dataset":
                fid.write("%s%s" % (linesep, linesep))
            fid.write("[%s]%s" % (section, linesep))
            if sections_comments is not None and section in sections_comments:
                dump_help(fid, sections_comments[section])

            for key, values in section_content.items():
                if options_level < _options_levels[values["type"]]:
                    continue
                if comments and values["help"].strip() != "":
                    dump_help(fid, values["help"])

                value = values["default"]
                if section in prefilled_values and key in prefilled_values[section]:
                    value = prefilled_values[section][key]
                if section in config and key in config[section]:
                    value = config[section][key]
                fid.write("%s = %s%s" % (key, value, linesep))


def _extract_nabuconfig_section(section, default_config):
    res = {}
    for key, val in default_config[section].items():
        res[key] = val["default"]
    return res


def _extract_nabuconfig_keyvals(default_config):
    res = {}
    for section in default_config.keys():
        res[section] = _extract_nabuconfig_section(section, default_config)
    return res


def get_default_nabu_config(default_config):
    """
    Return a dictionary with the default nabu configuration.
    """
    return _extract_nabuconfig_keyvals(default_config)


def _handle_modified_key(key, val, section, default_config, renamed_keys):
    if val is not None:
        return key, val, section
    if key in renamed_keys and renamed_keys[key]["section"] == section:
        info = renamed_keys[key]
        print(info["message"])
        print("This is deprecated since version %s and will result in an error in futures versions" % (info["since"]))
        section = info.get("new_section", section)
        if info["new_name"] == "":
            return None, None, section  # deleted key
        val = default_config[section].get(info["new_name"], None)
        return info["new_name"], val, section
    else:
        return key, None, section  # unhandled renamed/deleted key


def validate_config(config, default_config, renamed_keys, errors="warn"):
    """
    Validate a configuration dictionary against a "default" configuration dict.

    Parameters
    ----------
    config: dict
        configuration dict to be validated
    default_config: dict
        Reference configuration. Missing keys/sections from 'config' will be updated with
        keys from this dictionary.
    errors: str, optional
        What to do when an unknonw key/section is encountered. Possible actions are:
          - "warn": throw a warning, continue the validation
          - "raise": raise an error and exit
    """

    def error(msg):
        if errors == "raise":
            raise ValueError(msg)
        else:
            print("Error: %s" % msg)

    res_config = {}
    for section, section_content in config.items():
        # Ignore the "other" section
        if section.lower() == "other":
            continue
        if section not in default_config:
            error("Unknown section [%s]" % section)
            continue
        res_config[section] = _extract_nabuconfig_section(section, default_config)
        res_config[section].update(section_content)
        for key, value in res_config[section].items():
            opt = default_config[section].get(key, None)
            key, opt, section_updated = _handle_modified_key(key, opt, section, default_config, renamed_keys)
            if key is None:
                continue  # deleted key
            if opt is None:
                error("Unknown option '%s' in section [%s]" % (key, section_updated))
                continue
            validator = default_config[section_updated][key]["validator"]
            if section_updated not in res_config:  # missing section - handled later
                continue
            res_config[section_updated][key] = validator(section_updated, key, value)
    # Handle sections missing in config
    for section in set(default_config.keys()) - set(res_config.keys()):
        res_config[section] = _extract_nabuconfig_section(section, default_config)
        for key, value in res_config[section].items():
            validator = default_config[section][key]["validator"]
            res_config[section][key] = validator(section, key, value)
    return res_config


validate_nabu_config = deprecated("validate_nabu_config is renamed validate_config", do_print=True)(validate_config)


def overwrite_config(conf, overwritten_params):
    """
    Overwrite a (validated) configuration with a new parameters dict.

    Parameters
    ----------
    conf: dict
        Configuration dictionary, usually output from validate_config()
    overwritten_params: dict
        Configuration dictionary with the same layout, containing parameters to overwrite
    """
    overwritten_params = overwritten_params or {}
    for section, params in overwritten_params.items():
        if section not in conf:
            raise ValueError("Unknown section %s" % section)
        current_section = conf[section]
        for key in params.keys():
            if key not in current_section:
                raise ValueError("Unknown parameter '%s' in section '%s'" % (key, section))
            conf[section][key] = overwritten_params[section][key]
    # ---
    return conf



import numpy as np
from ..utils import docstring, get_cuda_srcfile, updiv
from ..cuda.processing import CudaProcessing, __has_pycuda__
from ..processing.padding_cuda import CudaPadding
from ..processing.fft_cuda import get_fft_class, get_available_fft_implems
from ..processing.transpose import CudaTranspose
from .rings import MunchDeringer, SinoMeanDeringer, VoDeringer

if __has_pycuda__:
    import pycuda.gpuarray as garray
    from ..cuda.kernel import CudaKernel

try:
    from pycudwt import Wavelets

    __have_pycudwt__ = True
except ImportError:
    __have_pycudwt__ = False

# pylint: disable=E0606


class CudaMunchDeringer(MunchDeringer):
    def __init__(
        self,
        sigma,
        sinos_shape,
        levels=None,
        wname="db15",
        padding=None,
        padding_mode="edge",
        fft_backend="vkfft",
        cuda_options=None,
    ):
        """
        Initialize a "Munch Et Al" sinogram deringer with the Cuda backend.
        See References for more information.

        Parameters
        -----------
        sigma: float
            Standard deviation of the damping parameter. The higher value of sigma,
            the more important the filtering effect on the rings.
        levels: int, optional
            Number of wavelets decomposition levels.
            By default (None), the maximum number of decomposition levels is used.
        wname: str, optional
            Default is "db15" (Daubechies, 15 vanishing moments)
        sinos_shape: tuple, optional
            Shape of the sinogram (or sinograms stack).

        References
        ----------
        B. Munch, P. Trtik, F. Marone, M. Stampanoni, Stripe and ring artifact removal with
        combined wavelet-Fourier filtering, Optics Express 17(10):8567-8591, 2009.
        """
        super().__init__(sigma, sinos_shape, levels=levels, wname=wname, padding=padding, padding_mode=padding_mode)
        self._check_can_use_wavelets()
        self.cuda_processing = CudaProcessing(**(cuda_options or {}))
        self.ctx = self.cuda_processing.ctx
        self._init_pycudwt()
        self._init_padding()
        self._init_fft(fft_backend)
        self._setup_fw_kernel()

    def _check_can_use_wavelets(self):
        if not (__have_pycudwt__ and __has_pycuda__):
            raise ValueError("Needs pycuda and pycudwt to use this class")

    def _init_padding(self):
        if self.padding is None:
            return
        self.padder = CudaPadding(
            self.sinos_shape[1:],
            ((0, 0), self.padding),
            mode=self.padding_mode,
            cuda_options={"ctx": self.cuda_processing.ctx},
        )

    def _init_fft(self, fft_backend):
        self.fft_cls = get_fft_class(backend=fft_backend)
        # For all k >= 1, we perform a batched (I)FFT along axis 0 on an array
        # of shape (n_a/2^k, n_x/2^k)  (up to DWT size rounding)
        if self.fft_cls.implem == "vkfft":
            self._create_plans_vkfft()
        else:
            self._create_plans_skfft()

    def _create_plans_skfft(self):
        self._fft_plans = {}
        for level, d_vcoeff in self._d_vertical_coeffs.items():
            self._fft_plans[level] = self.fft_cls(d_vcoeff.shape, np.float32, r2c=True, axes=(0,), ctx=self.ctx)

    def _create_plans_vkfft(self):
        """
        VKFFT does not support batched R2C transforms along axis 0 ("slow axis").
        We can either use C2C (faster, but needs more memory) or transpose the arrays to do R2C along axis=1.
        Here we transpose the arrays.
        """
        self._fft_plans = {}
        self._transpose_forward_1 = {}
        self._transpose_forward_2 = {}
        self._transpose_inverse_1 = {}
        self._transpose_inverse_2 = {}
        for level, d_vcoeff in self._d_vertical_coeffs.items():
            shape = d_vcoeff.shape
            # Normally, a batched 1D fft on 2D data of shape (Ny, Nx) along axis 0 returns an array of shape (Ny/2+1, Nx):
            #
            #  (Ny, Nx)  --[fft_0]--> (Ny/2, Nx)
            #    f32                      c64
            #
            # In this case, we can only do batched 1D transform along axis 1, so we have to trick with transposes:
            #
            #  (Ny, Nx) --[T]--> (Nx, Ny) --[fft_1]--> (Nx, Ny/2) --[T]--> (Ny/2, Nx)
            #    f32                f32                   c64                  c64
            #
            # (In both cases IFFT is done the same way from right to left)
            self._transpose_forward_1[level] = CudaTranspose(shape, np.float32, ctx=self.ctx)
            self._fft_plans[level] = self.fft_cls(shape[::-1], np.float32, r2c=True, ctx=self.ctx)
            self._transpose_forward_2[level] = CudaTranspose((shape[1], shape[0] // 2 + 1), np.complex64, ctx=self.ctx)
            self._transpose_inverse_1[level] = CudaTranspose((shape[0] // 2 + 1, shape[1]), np.complex64, ctx=self.ctx)
            self._transpose_inverse_2[level] = CudaTranspose(shape[::-1], np.float32, ctx=self.ctx)

    def _init_pycudwt(self):
        if self.levels is None:
            self.levels = 100  # will be clipped by pycudwt
        sino_shape = self.sinos_shape[1:] if self.padding is None else self.sino_padded_shape
        self.cudwt = Wavelets(np.zeros(sino_shape, "f"), self.wname, self.levels)
        self.levels = self.cudwt.levels
        # Access memory allocated by "pypwt" from pycuda
        self._d_sino = garray.empty(sino_shape, np.float32, gpudata=self.cudwt.image_int_ptr())
        self._get_vertical_coeffs()

    def _get_vertical_coeffs(self):
        self._d_vertical_coeffs = {}
        # Transfer the (0-memset) coefficients in order to get all the shapes
        coeffs = self.cudwt.coeffs
        for i in range(self.cudwt.levels):
            shape = coeffs[i + 1][1].shape
            self._d_vertical_coeffs[i + 1] = garray.empty(
                shape, np.float32, gpudata=self.cudwt.coeff_int_ptr(3 * i + 2)
            )

    def _setup_fw_kernel(self):
        self._fw_kernel = CudaKernel(
            "kern_fourierwavelets",
            filename=get_cuda_srcfile("fourier_wavelets.cu"),
            signature="Piif",
        )

    def _apply_fft(self, level):
        d_coeffs = self._d_vertical_coeffs[level]
        # All the memory is allocated (or re-used) under the hood
        if self.fft_cls.implem == "vkfft":
            d_coeffs_t = self._transpose_forward_1[level](
                d_coeffs
            )  # allocates self._transpose_forward_1[level].processing.dst
            d_coeffs_t_f = self._fft_plans[level].fft(d_coeffs_t)  # allocates self._fft_plans[level].output_fft
            d_coeffs_f = self._transpose_forward_2[level](
                d_coeffs_t_f
            )  # allocates self._transpose_forward_2[level].processing.dst
        else:
            d_coeffs_f = self._fft_plans[level].fft(d_coeffs)
        return d_coeffs_f

    def _apply_ifft(self, d_coeffs_f, level):
        d_coeffs = self._d_vertical_coeffs[level]
        if self.fft_cls.implem == "vkfft":
            d_coeffs_t_f = self._transpose_inverse_1[level](d_coeffs_f, dst=self._fft_plans[level].output_fft)
            d_coeffs_t = self._fft_plans[level].ifft(
                d_coeffs_t_f, output=self._transpose_forward_1[level].processing.dst
            )
            self._transpose_inverse_2[level](d_coeffs_t, dst=d_coeffs)
        else:
            self._fft_plans[level].ifft(d_coeffs_f, output=d_coeffs)

    def _destripe_2D(self, d_sino, output):
        if not (d_sino.flags.c_contiguous):
            sino = self.cuda_processing.allocate_array("_d_sino", d_sino.shape, np.float32)
            sino[:] = d_sino[:]
        else:
            sino = d_sino
        if self.padding is not None:
            sino = self.padder.pad(sino)
        # set the "image" for DWT (memcpy D2D)
        self._d_sino.set(sino)
        # perform forward DWT
        self.cudwt.forward()
        for i in range(self.cudwt.levels):
            level = i + 1
            Ny, Nx = self._d_vertical_coeffs[level].shape
            # Batched FFT along axis 0
            d_vertical_coeffs_f = self._apply_fft(level)

            # Dampen the wavelets coefficients
            self._fw_kernel(d_vertical_coeffs_f, Nx, Ny, self.sigma)
            # IFFT
            self._apply_ifft(d_vertical_coeffs_f, level)

        # Finally, inverse DWT
        self.cudwt.inverse()
        d_out = self._d_sino
        if self.padding is not None:
            d_out = self._d_sino[:, self.padding[0] : -self.padding[1]]  # memcpy2D
        output.set(d_out)
        return output


def can_use_cuda_deringer():
    """
    Check wether cuda implementation of deringer can be used.
    Checking for installed modules is not enough, as for example pyvkfft can be installed without cuda devices
    """
    can_do_fft = get_available_fft_implems() != []
    return can_do_fft and __have_pycudwt__


class CudaVoDeringer(VoDeringer):
    """
    An interface to topocupy's "remove_all_stripe".
    """

    def _check_requirement(self):
        # Do it here, otherwise cupy shows warnings at import even if not used
        from ..thirdparty.tomocupy_remove_stripe import remove_all_stripe_pycuda, __have_tomocupy_deringer__

        if not (__have_tomocupy_deringer__):
            raise ImportError("need cupy")
        self._remove_all_stripe_pycuda = remove_all_stripe_pycuda

    def remove_rings_radios(self, radios):
        return self._remove_all_stripe_pycuda(radios, layout="radios", **self._remove_all_stripe_kwargs)

    def remove_rings_sinograms(self, sinos):
        return self._remove_all_stripe_pycuda(sinos, layout="sinos", **self._remove_all_stripe_kwargs)

    def remove_rings_sinogram(self, sino):
        sinos = sino.reshape((1, sino.shape[0], -1))  # no copy
        self.remove_rings_sinograms(sinos)
        return sino

    remove_rings = remove_rings_sinograms


class CudaSinoMeanDeringer(SinoMeanDeringer):
    @docstring(SinoMeanDeringer)
    def __init__(
        self,
        sinos_shape,
        mode="subtract",
        filter_cutoff=None,
        padding_mode="edge",
        fft_num_threads=None,
        **cuda_options,
    ):
        self.processing = CudaProcessing(**(cuda_options or {}))
        super().__init__(sinos_shape, mode, filter_cutoff, padding_mode, fft_num_threads)
        self._init_kernels()

    def _init_kernels(self):
        self.d_sino_profile = self.processing.allocate_array("sino_profile", self.n_x)
        self._mean_kernel = self.processing.kernel(
            "vertical_mean",
            filename=get_cuda_srcfile("normalization.cu"),
            signature="PPiii",
        )
        self._mean_kernel_block = (32, 1, 32)
        self._mean_kernel_grid = [updiv(a, b) for a, b in zip(self.sinos_shape[::-1], self._mean_kernel_block)]
        self._mean_kernel_args = [self.d_sino_profile, np.int32(self.n_x), np.int32(self.n_angles), np.int32(self.n_z)]
        self._mean_kernel_kwargs = {
            "grid": self._mean_kernel_grid,
            "block": self._mean_kernel_block,
        }

        self._op_kernel = self.processing.kernel(
            "inplace_generic_op_3Dby1D",
            filename=get_cuda_srcfile("ElementOp.cu"),
            signature="PPiii",
            options=["-DGENERIC_OP=%d" % (3 if self.mode == "divide" else 1)],
        )
        self._op_kernel_block = (16, 16, 4)
        self._op_kernel_grid = [updiv(a, b) for a, b in zip(self.sinos_shape[::-1], self._op_kernel_block)]
        self._op_kernel_args = [self.d_sino_profile, np.int32(self.n_x), np.int32(self.n_angles), np.int32(self.n_z)]
        self._op_kernel_kwargs = {
            "grid": self._op_kernel_grid,
            "block": self._op_kernel_block,
        }

    def _init_filter(self, filter_cutoff, fft_num_threads, padding_mode):
        super()._init_filter(filter_cutoff, fft_num_threads, padding_mode)
        if filter_cutoff is None:
            return
        self._d_filter_f = self.processing.to_device("_filter_f", self._filter_f)

        self.padder = CudaPadding(
            (self.n_x, 1),
            ((self._pad_left, self._pad_right), (0, 0)),
            mode=self.padding_mode,
            cuda_options={"ctx": self.processing.ctx},
        )
        fft_cls = get_fft_class()
        self._fft = fft_cls(self._filter_size, np.float32, r2c=True)

    def _apply_filter(self, sino_profile):
        if self._filter_f is None:
            return sino_profile

        sino_profile = sino_profile.reshape((-1, 1))  # view
        sino_profile_p = self.padder.pad(sino_profile).ravel()

        sino_profile_f = self._fft.fft(sino_profile_p)
        sino_profile_f *= self._d_filter_f
        self._fft.ifft(sino_profile_f, output=sino_profile_p)

        self.d_sino_profile[:] = sino_profile_p[self._pad_left : -self._pad_right]
        return self.d_sino_profile

    def remove_rings_sinogram(self, sino, output=None):
        #
        if output is not None:
            raise NotImplementedError
        #
        if not (sino.flags.c_contiguous):
            d_sino = self.processing.allocate_array("d_sino", sino.shape, np.float32)
            d_sino[:] = sino[:]
        else:
            d_sino = sino
        self._mean_kernel(d_sino, *self._mean_kernel_args, **self._mean_kernel_kwargs)
        self._apply_filter(self.d_sino_profile)
        self._op_kernel(d_sino, *self._op_kernel_args, **self._op_kernel_kwargs)
        if not (sino.flags.c_contiguous):
            sino[:] = self.processing.d_sino[:]
        return sino

    def remove_rings_sinograms(self, sinograms):
        for i in range(sinograms.shape[0]):
            self.remove_rings_sinogram(sinograms[i])



from glob import glob
from pathlib import Path as pathlib_Path
from os import path, getcwd, chdir
from posixpath import join as posix_join
from datetime import datetime
import numpy as np
from h5py import VirtualSource, VirtualLayout
from silx.io.dictdump import dicttoh5
from silx.io.url import DataUrl

try:
    from tomoscan.io import HDF5File
except:
    from h5py import File as HDF5File
from tomoscan.esrf import RawVolume
from tomoscan.esrf.volume.jp2kvolume import has_glymur as __have_jp2k__
from .. import version as nabu_version
from ..utils import merged_shape
from .utils import convert_dict_values


def get_datetime():
    """
    Function used by some writers to indicate the current date.
    """
    return datetime.now().replace(microsecond=0).isoformat()


class Writer:
    """
    Base class for all writers.
    """

    def __init__(self, fname):
        self.fname = fname

    def get_filename(self):
        return self.fname


###################################################################################################
## Nabu original code for NXProcessWriter - also works for non-3D data, does not depend on tomoscan
###################################################################################################


def h5_write_object(h5group, key, value, overwrite=False, default_val=None):
    existing_val = h5group.get(key, default_val)
    if existing_val is not default_val:
        if not overwrite:
            raise OSError("Unable to create link (name already exists): %s" % h5group.name)
        else:
            h5group.pop(key)
    h5group[key] = value


class NXProcessWriter(Writer):
    """
    A class to write Nexus file with a processing result.
    """

    def __init__(self, fname, entry=None, filemode="a", overwrite=False):
        """
        Initialize a NXProcessWriter.

        Parameters
        -----------
        fname: str
            Path to the HDF5 file.
        entry: str, optional
            Entry in the HDF5 file. Default is "entry"
        """
        super().__init__(fname)
        self._set_entry(entry)
        self._filemode = filemode
        self.overwrite = overwrite

    def _set_entry(self, entry):
        self.entry = entry or "entry"
        data_path = posix_join("/", self.entry)
        self.data_path = data_path

    def write(
        self,
        result,
        process_name,
        processing_index=0,
        config=None,
        data_name="data",
        is_frames_stack=True,
        direct_access=True,
    ):
        """
        Write the result in the current NXProcess group.

        Parameters
        ----------
        result: numpy.ndarray
            Array containing the processing result
        process_name: str
            Name of the processing
        processing_index: int
            Index of the processing (in a pipeline)
        config: dict, optional
            Dictionary containing the configuration.
        """
        swmr = self._filemode == "r"
        with HDF5File(self.fname, self._filemode, swmr=swmr) as fid:
            nx_entry = fid.require_group(self.data_path)
            if "NX_class" not in nx_entry.attrs:
                nx_entry.attrs["NX_class"] = "NXentry"

            nx_process = nx_entry.require_group(process_name)
            nx_process.attrs["NX_class"] = "NXprocess"

            metadata = {
                "program": "nabu",
                "version": nabu_version,
                "date": get_datetime(),
                "sequence_index": np.int32(processing_index),
            }
            for key, val in metadata.items():
                h5_write_object(nx_process, key, val, overwrite=self.overwrite)

            if config is not None:
                export_dict_to_h5(
                    config, self.fname, posix_join(nx_process.name, "configuration"), overwrite_data=True, mode="a"
                )
                nx_process["configuration"].attrs["NX_class"] = "NXcollection"
            if isinstance(result, dict):
                results_path = posix_join(nx_process.name, "results")
                export_dict_to_h5(result, self.fname, results_path, overwrite_data=self.overwrite, mode="a")
            else:
                nx_data = nx_process.require_group("results")
                results_path = nx_data.name
                nx_data.attrs["NX_class"] = "NXdata"
                nx_data.attrs["signal"] = data_name

                results_data_path = posix_join(results_path, data_name)
                if self.overwrite and results_data_path in fid:
                    del fid[results_data_path]

                if isinstance(result, VirtualLayout):
                    nx_data.create_virtual_dataset(data_name, result)
                else:  # assuming array-like
                    nx_data[data_name] = result
                if is_frames_stack:
                    nx_data[data_name].attrs["interpretation"] = "image"
                nx_data.attrs["signal"] = data_name

            # prepare the direct access plots
            if direct_access:
                nx_process.attrs["default"] = "results"
                if "default" not in nx_entry.attrs:
                    nx_entry.attrs["default"] = posix_join(nx_process.name, "results")

            # Return the internal path to "results"
            return results_path


class NXVolVolume(NXProcessWriter):
    """
    An interface to NXProcessWriter with the same API than tomoscan.esrf.volume.

    NX files are written in two ways:

       1. Partial files containing sub-volumes
       2. Final volume: master file with virtual dataset pointing to partial files

    This class handles the first one, therefore expects the "start_index" parameter.
    In the case of HDF5, a sub-directory is creating to contain the partial files.
    In other words, if file_prefix="recons" and output_dir="/path/to/out":

    /path/to/out/recons.h5 # final master file
    /path/to/out/recons/
      /path/to/out/recons/recons_00000.h5
      /path/to/out/recons/recons_00100.h5
      ...
    """

    def __init__(self, **kwargs):
        # get parameters from kwargs passed to tomoscan XXVolume()
        folder = output_dir = kwargs.get("folder", None)
        volume_basename = file_prefix = kwargs.get("volume_basename", None)
        start_index = kwargs.get("start_index", None)
        overwrite = kwargs.get("overwrite", False)
        data_path = entry = kwargs.get("data_path", None)
        self._process_name = kwargs.get("process_name", "reconstruction")
        if any([param is None for param in [folder, volume_basename, start_index, entry]]):
            raise ValueError("Need the following parameters: folder, volume_basename, start_index, data_path")
        #

        # By default, a sub-folder is created so that partial volumes will be one folder below the master file
        # (see example above in class documentation)
        if kwargs.get("create_subfolder", True):
            output_dir = path.join(output_dir, file_prefix)

        if path.exists(output_dir):
            if not (path.isdir(output_dir)):
                raise ValueError("Unable to create directory %s: already exists and is not a directory" % output_dir)
        else:
            pathlib_Path(output_dir).mkdir(parents=True, exist_ok=True)
        #

        file_prefix += str("_%05d" % start_index)
        fname = path.join(output_dir, file_prefix + ".hdf5")

        super().__init__(fname, entry=entry, filemode="a", overwrite=overwrite)
        self.data = None
        self.metadata = None
        self.file_path = fname

    def save(self):
        if self.data is None:
            raise ValueError("Must set data first")
        self.write(self.data, self._process_name, config=self.metadata)

    def save_metadata(self):
        pass  # already done

    def browse_data_files(self):
        return [self.fname]


# COMPAT.
LegacyNXProcessWriter = NXProcessWriter
#

########################################################################################
########################################################################################
########################################################################################


def export_dict_to_h5(dic, h5file, h5path, overwrite_data=True, mode="a"):
    """
    Wrapper on top of silx.io.dictdump.dicttoh5 replacing None with "None"

    Parameters
    -----------
    dic: dict
        Dictionary containing the options
    h5file: str
        File name
    h5path: str
        Path in the HDF5 file
    overwrite_data: bool, optional
        Whether to overwrite data when writing HDF5. Default is True
    mode: str, optional
        File mode. Default is "a" (append).
    """
    modified_dic = convert_dict_values(
        dic,
        {None: "None"},
    )
    update_mode = {True: "modify", False: "add"}[bool(overwrite_data)]
    return dicttoh5(modified_dic, h5file=h5file, h5path=h5path, update_mode=update_mode, mode=mode)


def create_virtual_layout(files_or_pattern, h5_path, base_dir=None, axis=0, dtype="f"):
    """
    Create a HDF5 virtual layout.

    Parameters
    ----------
    files_or_pattern: str or list
        A list of file names, or a wildcard pattern.
        If a list is provided, it will not be sorted! This will have to be
        done before calling this function.
    h5_path: str
        Path inside the HDF5 input file(s)
    base_dir: str, optional
        Base directory when using relative file names.
    axis: int, optional
        Data axis to merge. Default is 0.
    """
    prev_cwd = None
    if base_dir is not None:
        prev_cwd = getcwd()
        chdir(base_dir)
    if isinstance(files_or_pattern, str):
        files_list = glob(files_or_pattern)
        files_list.sort()
    else:  # list
        files_list = files_or_pattern
    if files_list == []:
        raise ValueError("Nothing found as pattern %s" % files_or_pattern)
    virtual_sources = []
    shapes = []
    for fname in files_list:
        with HDF5File(fname, "r", swmr=True) as fid:
            shape = fid[h5_path].shape
        vsource = VirtualSource(fname, name=h5_path, shape=shape)
        virtual_sources.append(vsource)
        shapes.append(shape)
    total_shape = merged_shape(shapes, axis=axis)

    virtual_layout = VirtualLayout(shape=total_shape, dtype=dtype)
    start_idx = 0
    for vsource, shape in zip(virtual_sources, shapes):
        n_imgs = shape[axis]
        # Perhaps there is more elegant
        if axis == 0:
            virtual_layout[start_idx : start_idx + n_imgs] = vsource
        elif axis == 1:
            virtual_layout[:, start_idx : start_idx + n_imgs, :] = vsource
        elif axis == 2:
            virtual_layout[:, :, start_idx : start_idx + n_imgs] = vsource
        else:
            raise ValueError("Only axis 0,1,2 are supported")
        #
        start_idx += n_imgs

    if base_dir is not None:
        chdir(prev_cwd)
    return virtual_layout


def merge_hdf5_files(
    files_or_pattern,
    h5_path,
    output_file,
    process_name,
    output_entry=None,
    output_filemode="a",
    data_name="data",
    processing_index=0,
    config=None,
    base_dir=None,
    axis=0,
    overwrite=False,
    dtype="f",
):
    """
    Parameters
    -----------
    files_or_pattern: str or list
        A list of file names, or a wildcard pattern.
        If a list is provided, it will not be sorted! This will have to be
        done before calling this function.
    h5_path: str
        Path inside the HDF5 input file(s)
    output_file: str
        Path of the output file
    process_name: str
        Name of the process
    output_entry: str, optional
        Output HDF5 root entry (default is "/entry")
    output_filemode: str, optional
        File mode for output file. Default is "a" (append)
    processing_index: int, optional
        Processing index for the output file. Default is 0.
    config: dict, optional
        Dictionary describing the configuration needed to get the results.
    base_dir: str, optional
        Base directory when using relative file names.
    axis: int, optional
        Data axis to merge. Default is 0.
    overwrite: bool, optional
        Whether to overwrite already existing data in the final file.
        Default is False.
    """
    if base_dir is not None:
        prev_cwd = getcwd()
    virtual_layout = create_virtual_layout(files_or_pattern, h5_path, base_dir=base_dir, axis=axis, dtype=dtype)
    nx_file = NXProcessWriter(output_file, entry=output_entry, filemode=output_filemode, overwrite=overwrite)
    nx_file.write(
        virtual_layout,
        process_name,
        processing_index=processing_index,
        config=config,
        data_name=data_name,
        is_frames_stack=True,
    )
    # pylint: disable=E0606
    if base_dir is not None and prev_cwd != getcwd():
        chdir(prev_cwd)


class HSTVolWriter(Writer):
    """
    A writer to mimic PyHST2 ".vol" files
    """

    def __init__(self, fname, append=False, **kwargs):
        super().__init__(fname)
        self.append = append
        self._vol_writer = RawVolume(fname, overwrite=True, append=append)
        self._hst_metadata = kwargs.get("hst_metadata", {})

    def generate_metadata(self, data, **kwargs):
        n_z, n_y, n_x = data.shape
        metadata = {
            "NUM_X": n_x,
            "NUM_Y": n_y,
            "NUM_Z": n_z,
            "voxelSize": 40.0,
            "BYTEORDER": "LOWBYTEFIRST",
            "ValMin": kwargs.get("ValMin", 0.0),
            "ValMax": kwargs.get("ValMin", 1.0),
            "s1": 0.0,
            "s2": 0.0,
            "S1": 0.0,
            "S2": 0.0,
        }
        for key, default_val in metadata.items():
            metadata[key] = kwargs.get(key, None) or self._hst_metadata.get(key, None) or default_val
        return metadata

    @staticmethod
    def sanitize_metadata(metadata):
        # To be fixed in RawVolume
        for what in ["NUM_X", "NUM_Y", "NUM_Z"]:
            metadata[what] = int(metadata[what])
        for what in ["voxelSize", "ValMin", "ValMax", "s1", "s2", "S1", "S2"]:
            metadata[what] = float(metadata[what])

    def write(self, data, *args, config=None, **kwargs):
        existing_metadata = self._vol_writer.load_metadata()
        new_metadata = self.generate_metadata(data)
        if len(existing_metadata) == 0 or not (self.append):
            # first write or append==False
            metadata = new_metadata
        else:
            # append write ; update metadata
            metadata = existing_metadata.copy()
            self.sanitize_metadata(metadata)
            metadata["NUM_Z"] += new_metadata["NUM_Z"]
        self._vol_writer.data = data
        self._vol_writer.metadata = metadata
        self._vol_writer.save()
        # Also save .xml
        self._vol_writer.save_metadata(
            url=DataUrl(
                scheme="lxml",
                file_path=self._vol_writer.metadata_url.file_path().replace(".info", ".xml"),
            )
        )


class HSTVolVolume(HSTVolWriter):
    """
    An interface to HSTVolWriter with the same API than tomoscan.esrf.volume.
    This is really not ideal, see nabu:#381
    """

    def __init__(self, **kwargs):
        file_path = kwargs.get("file_path", None)
        if file_path is None:
            raise ValueError("Missing mandatory 'file_path' parameter")
        super().__init__(file_path, append=kwargs.pop("append", False), **kwargs)
        self.data = None
        self.metadata = None
        self.data_url = self._vol_writer.data_url

    def save(self):
        if self.data is None:
            raise ValueError("Must set data first")
        self.write(self.data)

    def save_metadata(self):
        pass  # already done for HST part - proper metadata is not supported

    def browse_data_files(self):
        return [self.fname]



import os
import numpy as np
import pytest
import scipy.ndimage
import h5py
from nabu.testutils import utilstest, __do_long_tests__
from nabu.testutils import get_data

from nabu.estimation.cor import (
    CenterOfRotation,
    CenterOfRotationAdaptiveSearch,
    CenterOfRotationGrowingWindow,
    CenterOfRotationSlidingWindow,
    CenterOfRotationOctaveAccurate,
)
from nabu.estimation.cor_sino import SinoCor, CenterOfRotationFourierAngles, CenterOfRotationVo, __have_algotom__


@pytest.fixture(scope="class")
def bootstrap_cor(request):
    cls = request.cls
    cls.abs_tol = 0.2
    cls.data, calib_data = get_cor_data_h5("test_alignment_cor.h5")
    cls.cor_gl_pix, cls.cor_hl_pix, cls.tilt_deg = calib_data


@pytest.fixture(scope="class")
def bootstrap_cor_win(request):
    cls = request.cls
    cls.abs_tol = 0.2
    cls.data_ha_proj, cls.cor_ha_pr_pix = get_cor_win_proj_data_h5("ha_autocor_radios.npz")
    cls.data_ha_sino, cls.cor_ha_sn_pix = get_cor_win_sino_data_h5("halftomo_1_sino.npz")


@pytest.fixture(scope="class")
def bootstrap_cor_accurate(request):
    cls = request.cls
    cls.abs_tol = 0.2
    cls.image_pair_stylo, cls.cor_pos_abs_stylo = get_cor_win_proj_data_h5("stylo_accurate.npz")
    cls.image_pair_blc12781, cls.cor_pos_abs_blc12781 = get_cor_win_proj_data_h5("blc12781_accurate.npz")


@pytest.fixture(scope="class")
def bootstrap_cor_fourier(request):
    cls = request.cls
    cls.abs_tol = 0.2
    dataset_relpath = os.path.join("sino_bamboo_hercules_for_test.npz")
    dataset_downloaded_path = utilstest.getfile(dataset_relpath)
    a = np.load(dataset_downloaded_path)
    cls.sinos = a["sinos"]
    cls.angles = a["angles"]
    cls.true_cor = a["true_cor"]


def get_cor_data_h5(*dataset_path):
    """
    Get a dataset file from silx.org/pub/nabu/data
    dataset_args is a list describing a nested folder structures, ex.
    ["path", "to", "my", "dataset.h5"]
    """
    dataset_relpath = os.path.join(*dataset_path)
    dataset_downloaded_path = utilstest.getfile(dataset_relpath)
    with h5py.File(dataset_downloaded_path, "r") as hf:
        data = hf["/entry/instrument/detector/data"][()]
        cor_global_pix = hf["/calibration/alignment/global/x_rotation_axis_pixel_position"][()][0]
        cor_highlow_pix = hf["/calibration/alignment/highlow/x_rotation_axis_pixel_position"][()][0]
        tilt_deg = hf["/calibration/alignment/highlow/z_camera_tilt"][()][0]

    return data, (cor_global_pix, cor_highlow_pix, tilt_deg)


def get_cor_win_proj_data_h5(*dataset_path):
    """
    Get a dataset file from silx.org/pub/nabu/data
    dataset_args is a list describing a nested folder structures, ex.
    ["path", "to", "my", "dataset.h5"]
    """
    dataset_relpath = os.path.join(*dataset_path)
    dataset_downloaded_path = utilstest.getfile(dataset_relpath)
    data = np.load(dataset_downloaded_path)
    radios = np.stack((data["radio1"], data["radio2"]), axis=0)

    return radios, data["cor_pos"]


def get_cor_win_sino_data_h5(*dataset_path):
    """
    Get a dataset file from silx.org/pub/nabu/data
    dataset_args is a list describing a nested folder structures, ex.
    ["path", "to", "my", "dataset.h5"]
    """
    dataset_relpath = os.path.join(*dataset_path)
    dataset_downloaded_path = utilstest.getfile(dataset_relpath)
    data = np.load(dataset_downloaded_path)
    sino_shape = data["sino"].shape
    sinos = np.stack((data["sino"][: sino_shape[0] // 2], data["sino"][sino_shape[0] // 2 :]), axis=0)

    return sinos, data["cor"] - sino_shape[1] / 2


@pytest.mark.usefixtures("bootstrap_cor")
class TestCor:
    def test_cor_posx(self):
        radio1 = self.data[0, :, :]
        radio2 = np.fliplr(self.data[1, :, :])

        CoR_calc = CenterOfRotation()
        cor_position = CoR_calc.find_shift(radio1, radio2, return_relative_to_middle=True)
        assert np.isscalar(cor_position), f"cor_position expected to be a scale, {type(cor_position)} returned"

        message = "Computed CoR %f " % cor_position + " and real CoR %f do not coincide" % self.cor_gl_pix
        assert np.isclose(self.cor_gl_pix, cor_position, atol=self.abs_tol), message

        # testing again with the validity return value
        cor_position, result_validity = CoR_calc.find_shift(
            radio1, radio2, return_validity=True, return_relative_to_middle=True
        )
        assert np.isscalar(cor_position)

        message = (
            "returned result_validity is  %s " % result_validity
            + " while it should be unknown because the validity check is not yet implemented"
        )
        assert result_validity == "unknown", message

    def test_noisy_cor_posx(self):
        radio1 = np.fmax(self.data[0, :, :], 0)
        radio2 = np.fmax(self.data[1, :, :], 0)

        radio1 = np.random.poisson(radio1 * 400)
        radio2 = np.random.poisson(np.fliplr(radio2) * 400)

        CoR_calc = CenterOfRotation()
        cor_position = CoR_calc.find_shift(radio1, radio2, median_filt_shape=(3, 3), return_relative_to_middle=True)
        assert np.isscalar(cor_position), f"cor_position expected to be a scale, {type(cor_position)} returned"

        message = "Computed CoR %f " % cor_position + " and real CoR %f do not coincide" % self.cor_gl_pix
        assert np.isscalar(cor_position)
        assert np.isclose(self.cor_gl_pix, cor_position, atol=self.abs_tol), message

    def test_noisyHF_cor_posx(self):
        """test with noise at high frequencies"""
        radio1 = self.data[0, :, :]
        radio2 = np.fliplr(self.data[1, :, :])

        noise_level = radio1.max() / 16.0
        noise_ima1 = np.random.normal(0.0, size=radio1.shape) * noise_level
        noise_ima2 = np.random.normal(0.0, size=radio2.shape) * noise_level

        noise_ima1 = noise_ima1 - scipy.ndimage.gaussian_filter(noise_ima1, 2.0)
        noise_ima2 = noise_ima2 - scipy.ndimage.gaussian_filter(noise_ima2, 2.0)

        radio1 = radio1 + noise_ima1
        radio2 = radio2 + noise_ima2

        CoR_calc = CenterOfRotation()

        cor_position = CoR_calc.find_shift(radio1, radio2, low_pass=(6.0, 0.3), return_relative_to_middle=True)
        assert np.isscalar(cor_position), f"cor_position expected to be a scale, {type(cor_position)} returned"

        message = "Computed CoR %f " % cor_position + " and real CoR %f do not coincide" % self.cor_gl_pix
        assert np.isclose(self.cor_gl_pix, cor_position, atol=self.abs_tol), message

    @pytest.mark.skipif(not (__do_long_tests__), reason="need environment variable NABU_LONG_TESTS=1")
    def test_half_tomo_cor_exp(self):
        """test the half_tomo algorithm on experimental data"""

        radios = get_data("ha_autocor_radios.npz")
        radio1 = radios["radio1"]
        radio2 = radios["radio2"]
        cor_pos = radios["cor_pos"]

        radio2 = np.fliplr(radio2)

        CoR_calc = CenterOfRotationAdaptiveSearch()

        cor_position = CoR_calc.find_shift(
            radio1, radio2, low_pass=1, high_pass=20, filtered_cost=True, return_relative_to_middle=True
        )
        assert np.isscalar(cor_position), f"cor_position expected to be a scale, {type(cor_position)} returned"

        message = (
            "Computed CoR %f " % cor_position
            + " and real CoR %f should coincide when using the halftomo algorithm with half tomo data" % cor_pos
        )
        assert np.isclose(cor_pos, cor_position, atol=self.abs_tol + 0.5), message

    @pytest.mark.skipif(not (__do_long_tests__), reason="need environment variable NABU_LONG_TESTS=1")
    def test_half_tomo_cor_exp_limited(self):
        """test the hal_tomo algorithm on experimental data and global search with limits"""

        radios = get_data("ha_autocor_radios.npz")
        radio1 = radios["radio1"]
        radio2 = radios["radio2"]
        cor_pos = radios["cor_pos"]

        radio2 = np.fliplr(radio2)

        CoR_calc = CenterOfRotationAdaptiveSearch()

        cor_position, result_validity = CoR_calc.find_shift(
            radio1,
            radio2,
            low_pass=1,
            high_pass=20,
            margins=(100, 10),
            filtered_cost=False,
            return_validity=True,
            return_relative_to_middle=True,
        )
        assert np.isscalar(cor_position), f"cor_position expected to be a scale, {type(cor_position)} returned"

        message = (
            "Computed CoR %f " % cor_position
            + " and real CoR %f should coincide when using the halftomo algorithm with half tomo data" % cor_pos
        )
        assert np.isclose(cor_pos, cor_position, atol=self.abs_tol + 0.5), message

        message = "returned result_validity is  %s " % result_validity + " while it should be sound"

        assert result_validity == "sound", message

    def test_cor_posx_linear(self):
        radio1 = self.data[0, :, :]
        radio2 = np.fliplr(self.data[1, :, :])

        CoR_calc = CenterOfRotation()
        cor_position = CoR_calc.find_shift(radio1, radio2, padding_mode="edge", return_relative_to_middle=True)
        assert np.isscalar(cor_position), f"cor_position expected to be a scale, {type(cor_position)} returned"

        message = "Computed CoR %f " % cor_position + " and real CoR %f do not coincide" % self.cor_gl_pix
        assert np.isclose(self.cor_gl_pix, cor_position, atol=self.abs_tol), message

    def test_error_checking_001(self):
        CoR_calc = CenterOfRotation()

        radio1 = self.data[0, :, :1:]
        radio2 = self.data[1, :, :]

        with pytest.raises(ValueError) as ex:
            CoR_calc.find_shift(radio1, radio2, return_relative_to_middle=True)

        message = "Error should have been raised about img #1 shape, other error raised instead:\n%s" % str(ex.value)
        assert "Images need to be 2-dimensional. Shape of image #1" in str(ex.value), message

    def test_error_checking_002(self):
        CoR_calc = CenterOfRotation()

        radio1 = self.data[0, :, :]
        radio2 = self.data

        with pytest.raises(ValueError) as ex:
            CoR_calc.find_shift(radio1, radio2, return_relative_to_middle=True)

        message = "Error should have been raised about img #2 shape, other error raised instead:\n%s" % str(ex.value)
        assert "Images need to be 2-dimensional. Shape of image #2" in str(ex.value), message

    def test_error_checking_003(self):
        CoR_calc = CenterOfRotation()

        radio1 = self.data[0, :, :]
        radio2 = self.data[1, :, 0:10]

        with pytest.raises(ValueError) as ex:
            CoR_calc.find_shift(radio1, radio2, return_relative_to_middle=True)

        message = (
            "Error should have been raised about different image shapes, "
            + "other error raised instead:\n%s" % str(ex.value)
        )
        assert "Images need to be of the same shape" in str(ex.value), message


@pytest.mark.skipif(not (__do_long_tests__), reason="Need NABU_LONG_TESTS=1 for this test")
@pytest.mark.usefixtures("bootstrap_cor", "bootstrap_cor_win")
class TestCorWindowSlide:
    def test_proj_center_axis_lft(self):
        radio1 = self.data[0, :, :]
        radio2 = np.fliplr(self.data[1, :, :])

        CoR_calc = CenterOfRotationSlidingWindow()
        cor_position = CoR_calc.find_shift(
            radio1,
            radio2,
            side="left",
            window_width=round(radio1.shape[-1] / 4.0 * 3.0),
            return_relative_to_middle=True,
        )
        assert np.isscalar(cor_position), f"cor_position expected to be a scale, {type(cor_position)} returned"

        message = "Computed CoR %f " % cor_position + " and expected CoR %f do not coincide" % self.cor_gl_pix
        assert np.isclose(self.cor_gl_pix, cor_position, atol=self.abs_tol), message

        cor_position, result_validity = CoR_calc.find_shift(
            radio1,
            radio2,
            side="left",
            window_width=round(radio1.shape[-1] / 4.0 * 3.0),
            return_validity=True,
            return_relative_to_middle=True,
        )

        message = "returned result_validity is  %s " % result_validity + " while it should be sound"

        assert result_validity == "sound", message

    def test_proj_center_axis_cen(self):
        radio1 = self.data[0, :, :]
        radio2 = np.fliplr(self.data[1, :, :])

        CoR_calc = CenterOfRotationSlidingWindow()
        cor_position = CoR_calc.find_shift(radio1, radio2, side="center", return_relative_to_middle=True)
        assert np.isscalar(cor_position), f"cor_position expected to be a scale, {type(cor_position)} returned"

        message = "Computed CoR %f " % cor_position + " and expected CoR %f do not coincide" % self.cor_gl_pix
        assert np.isclose(self.cor_gl_pix, cor_position, atol=self.abs_tol), message

    def test_proj_right_axis_rgt(self):
        radio1 = self.data_ha_proj[0, :, :]
        radio2 = np.fliplr(self.data_ha_proj[1, :, :])

        CoR_calc = CenterOfRotationSlidingWindow()
        cor_position = CoR_calc.find_shift(radio1, radio2, side="right", return_relative_to_middle=True)
        assert np.isscalar(cor_position), f"cor_position expected to be a scale, {type(cor_position)} returned"

        message = "Computed CoR %f " % cor_position + " and expected CoR %f do not coincide" % self.cor_ha_pr_pix
        assert np.isclose(self.cor_ha_pr_pix, cor_position, atol=self.abs_tol), message

    def test_proj_left_axis_lft(self):
        radio1 = np.fliplr(self.data_ha_proj[0, :, :])
        radio2 = self.data_ha_proj[1, :, :]

        CoR_calc = CenterOfRotationSlidingWindow()
        cor_position = CoR_calc.find_shift(radio1, radio2, side="left", return_relative_to_middle=True)
        assert np.isscalar(cor_position), f"cor_position expected to be a scale, {type(cor_position)} returned"

        message = "Computed CoR %f " % cor_position + " and expected CoR %f do not coincide" % -self.cor_ha_pr_pix
        assert np.isclose(-self.cor_ha_pr_pix, cor_position, atol=self.abs_tol), message

    def test_sino_right_axis_rgt(self):
        sino1 = self.data_ha_sino[0, :, :]
        sino2 = np.fliplr(self.data_ha_sino[1, :, :])

        CoR_calc = CenterOfRotationSlidingWindow()
        cor_position = CoR_calc.find_shift(sino1, sino2, side="right", return_relative_to_middle=True)
        assert np.isscalar(cor_position), f"cor_position expected to be a scale, {type(cor_position)} returned"

        message = "Computed CoR %f " % cor_position + " and expected CoR %f do not coincide" % self.cor_ha_sn_pix
        assert np.isclose(self.cor_ha_sn_pix, cor_position, atol=self.abs_tol * 5), message


@pytest.mark.skipif(not (__do_long_tests__), reason="need NABU_LONG_TESTS for this test")
@pytest.mark.usefixtures("bootstrap_cor", "bootstrap_cor_win")
class TestCorWindowGrow:
    def test_proj_center_axis_cen(self):
        radio1 = self.data[0, :, :]
        radio2 = np.fliplr(self.data[1, :, :])

        CoR_calc = CenterOfRotationGrowingWindow()
        cor_position = CoR_calc.find_shift(radio1, radio2, side="center", return_relative_to_middle=True)
        assert np.isscalar(cor_position), f"cor_position expected to be a scale, {type(cor_position)} returned"
        message = "Computed CoR %f " % cor_position + " and expected CoR %f do not coincide" % self.cor_gl_pix
        assert np.isclose(self.cor_gl_pix, cor_position, atol=self.abs_tol), message

    def test_proj_right_axis_rgt(self):
        radio1 = self.data_ha_proj[0, :, :]
        radio2 = np.fliplr(self.data_ha_proj[1, :, :])

        CoR_calc = CenterOfRotationGrowingWindow()
        cor_position = CoR_calc.find_shift(radio1, radio2, side="right", return_relative_to_middle=True)
        assert np.isscalar(cor_position), f"cor_position expected to be a scale, {type(cor_position)} returned"

        message = "Computed CoR %f " % cor_position + " and expected CoR %f do not coincide" % self.cor_ha_pr_pix
        assert np.isclose(self.cor_ha_pr_pix, cor_position, atol=self.abs_tol), message

    def test_proj_left_axis_lft(self):
        radio1 = np.fliplr(self.data_ha_proj[0, :, :])
        radio2 = self.data_ha_proj[1, :, :]

        CoR_calc = CenterOfRotationGrowingWindow()
        cor_position = CoR_calc.find_shift(radio1, radio2, side="left", return_relative_to_middle=True)
        assert np.isscalar(cor_position), f"cor_position expected to be a scale, {type(cor_position)} returned"

        message = "Computed CoR %f " % cor_position + " and expected CoR %f do not coincide" % -self.cor_ha_pr_pix
        assert np.isclose(-self.cor_ha_pr_pix, cor_position, atol=self.abs_tol), message

        cor_position, result_validity = CoR_calc.find_shift(
            radio1, radio2, side="left", return_validity=True, return_relative_to_middle=True
        )

        message = "returned result_validity is  %s " % result_validity + " while it should be sound"

        assert result_validity == "sound", message

    def test_proj_right_axis_all(self):
        radio1 = self.data_ha_proj[0, :, :]
        radio2 = np.fliplr(self.data_ha_proj[1, :, :])

        CoR_calc = CenterOfRotationGrowingWindow()
        cor_position = CoR_calc.find_shift(radio1, radio2, side="all", return_relative_to_middle=True)
        assert np.isscalar(cor_position), f"cor_position expected to be a scale, {type(cor_position)} returned"

        message = "Computed CoR %f " % cor_position + " and expected CoR %f do not coincide" % self.cor_ha_pr_pix
        assert np.isclose(self.cor_ha_pr_pix, cor_position, atol=self.abs_tol), message

    def test_sino_right_axis_rgt(self):
        sino1 = self.data_ha_sino[0, :, :]
        sino2 = np.fliplr(self.data_ha_sino[1, :, :])

        CoR_calc = CenterOfRotationGrowingWindow()
        cor_position = CoR_calc.find_shift(sino1, sino2, side="right", return_relative_to_middle=True)
        assert np.isscalar(cor_position), f"cor_position expected to be a scale, {type(cor_position)} returned"

        message = "Computed CoR %f " % cor_position + " and expected CoR %f do not coincide" % self.cor_ha_sn_pix
        assert np.isclose(self.cor_ha_sn_pix, cor_position, atol=self.abs_tol * 4), message


@pytest.mark.usefixtures("bootstrap_cor_win")
class TestCoarseToFineSinoCor:
    def test_coarse_to_fine(self):
        """
        Test nabu.estimation.cor_sino.SinoCor
        """
        sino_halftomo = np.vstack([self.data_ha_sino[0], self.data_ha_sino[1]])
        sino_cor = SinoCor(self.data_ha_sino[0], np.fliplr(self.data_ha_sino[1]))
        cor_coarse = sino_cor.estimate_cor_coarse()
        assert np.isscalar(cor_coarse), f"cor_position expected to be a scalar, {type(cor_coarse)} returned"
        cor_fine = sino_cor.estimate_cor_fine()
        assert np.isscalar(cor_fine), f"cor_position expected to be a scale, {type(cor_fine)} returned"

        cor_ref = self.cor_ha_sn_pix + sino_halftomo.shape[-1] / 2.0
        message = "Computed CoR %f " % cor_fine + " and expected CoR %f do not coincide" % cor_ref
        assert abs(cor_fine - cor_ref) < self.abs_tol * 2, message


@pytest.mark.usefixtures("bootstrap_cor_accurate")
class TestCorOctaveAccurate:
    def test_cor_accurate_positive_shift(self):
        detector_width = self.image_pair_stylo[0].shape[1]
        CoR_calc = CenterOfRotationOctaveAccurate()
        cor_position = CoR_calc.find_shift(
            self.image_pair_stylo[0], np.fliplr(self.image_pair_stylo[1]), "center", return_relative_to_middle=True
        )
        cor_position = cor_position + detector_width / 2
        assert np.isscalar(cor_position), f"cor_position expected to be a scalar, {type(cor_position)} returned"
        message = f"Computed CoR {cor_position} and expected CoR {self.cor_pos_abs_stylo} do not coincide."
        assert np.isclose(self.cor_pos_abs_stylo, cor_position, atol=self.abs_tol), message

    def test_cor_accurate_negative_shift(self):
        detector_width = self.image_pair_blc12781[0].shape[1]
        CoR_calc = CenterOfRotationOctaveAccurate()
        cor_position = CoR_calc.find_shift(
            self.image_pair_blc12781[0],
            np.fliplr(self.image_pair_blc12781[1]),
            "center",
            return_relative_to_middle=True,
        )
        cor_position = cor_position + detector_width / 2
        assert np.isscalar(cor_position), f"cor_position expected to be a scalar, {type(cor_position)} returned"
        message = f"Computed CoR {cor_position} and expected CoR {self.cor_pos_abs_blc12781} do not coincide."
        assert np.isclose(self.cor_pos_abs_blc12781, cor_position, atol=self.abs_tol), message


@pytest.mark.usefixtures("bootstrap_cor_fourier", "bootstrap_cor_win")
class TestCorFourierAngle:
    @pytest.mark.skip("Broken function")
    def test_sino_right_axis_with_near_pos(self):
        sino = np.vstack([self.data_ha_sino[0], self.data_ha_sino[1]])
        start_angle = np.pi / 4
        angles = np.linspace(start_angle, start_angle + 2 * np.pi, sino.shape[0])

        CoR_calc = CenterOfRotationFourierAngles()
        cor_position = CoR_calc.find_shift(
            sino, angles, side="right", crop_around_cor=True, return_relative_to_middle=True
        )  # side=sino.shape[1]/2+740)
        assert np.isscalar(cor_position), f"cor_position expected to be a scale, {type(cor_position)} returned"

        message = "Computed CoR %f " % cor_position + " and expected CoR %f do not coincide" % self.cor_ha_sn_pix
        assert np.isclose(self.cor_ha_sn_pix, cor_position, atol=self.abs_tol * 3), message

    def test_sino_right_axis_with_near_pos_jl(self):

        CoR_calc = CenterOfRotationFourierAngles()
        cor_position = CoR_calc.find_shift(
            self.sinos, self.angles, side="right", crop_around_cor=True, return_relative_to_middle=True
        )  # side=sino.shape[1]/2+740)
        assert np.isscalar(cor_position), f"cor_position expected to be a scale, {type(cor_position)} returned"

        message = "Computed CoR %f " % cor_position + " and expected CoR %f do not coincide" % self.true_cor
        assert np.isclose(self.true_cor, cor_position, atol=self.abs_tol * 3), message


@pytest.fixture(scope="class")
def bootstrap_vo_cor(request):
    cls = request.cls
    cls.tol = 1e-2
    cls.test_sinograms = {name: get_data("sino_%s.npz" % name) for name in ["pencil", "coffee", "mousebrains"]}
    sino_bamboo = get_data("sino_bamboo_hercules_for_test.npz")
    cls.test_sinograms["bamboo_hercules"] = {
        "data": sino_bamboo["sinos"],
        # FIXME the test file needs to be re-generated, "true_cor" has an incorrect offset
        "cor": sino_bamboo["true_cor"] + (2560) / 2,
    }


@pytest.mark.skipif(not (__have_algotom__), reason="need algotom for this test")
@pytest.mark.usefixtures("bootstrap_vo_cor")
class TestVoCOR:

    def _test_cor(self, dataset_name, tolerance=1e-2, **cor_options):
        cor_finder = CenterOfRotationVo()
        cor = cor_finder.find_shift(
            self.test_sinograms[dataset_name]["data"], return_relative_to_middle=False, **cor_options
        )
        cor_ref = self.test_sinograms[dataset_name]["cor"]
        assert (
            np.abs(cor - cor_ref) < tolerance
        ), "CoR estimation failed for %s: expected %.3f, got %.3f (tol = %.2e)" % (
            dataset_name,
            cor_ref,
            cor,
            tolerance,
        )

    def test_cor_180(self):
        self._test_cor("pencil", tolerance=0.6)

    def test_cor_180_more_complex(self): ...

    def test_cor_360_halftomo(self):
        self._test_cor("bamboo_hercules", tolerance=0.1, halftomo=True)

    def test_cor_360_halftomo_hard(self):
        # This one is difficult
        self._test_cor("mousebrains", tolerance=2, halftomo=True)

    def test_cor_360_not_halftomo(self):
        self._test_cor("coffee", tolerance=0.5, halftomo=False, is_360=True)



from .alignment import AlignmentBase
from .cor import (
    CenterOfRotation,
    CenterOfRotationSlidingWindow,
    CenterOfRotationGrowingWindow,
    CenterOfRotationAdaptiveSearch,
)
from .cor_sino import SinoCor
from .distortion import estimate_flat_distortion
from .focus import CameraFocus
from .tilt import CameraTilt
from .translation import DetectorTranslationAlongBeam



from ..processing.convolution_cuda import *
from ..utils import deprecation_warning

deprecation_warning(
    "nabu.cuda.convolution has been moved to nabu.processing.convolution_cuda",
    do_print=True,
    func_name="convolution_cuda",
)



import pyopencl.array as parray
from pyopencl import Program, CommandQueue, kernel_work_group_info
from ..utils import (
    deprecation_warning,
    catch_warnings,
)  # TODO use warnings.catch_warnings once python < 3.11 is dropped
from ..processing.kernel_base import KernelBase


class OpenCLKernel(KernelBase):
    """
    Helper class that wraps OpenCL kernel through pyopencl.

    Parameters
    -----------
    kernel_name: str
        Name of the OpenCL kernel.
    ctx: pyopencl.Context
        OpenCL context to use.
    queue: pyopencl.CommandQueue
        OpenCL queue to use. If provided, will use this queue's context instead of 'ctx'
    filename: str, optional
        Path to the file name containing kernels definitions
    src: str, optional
        Source code of kernels definitions
    automation_params: dict, optional
        Automation parameters, see below
    build_kwargs: optional
        Extra arguments to provide to pyopencl.Program.build(),
    """

    def __init__(
        self,
        kernel_name,
        ctx,
        queue=None,
        filename=None,
        src=None,
        automation_params=None,
        silent_compilation_warnings=False,
        **build_kwargs,
    ):
        super().__init__(
            kernel_name,
            filename=filename,
            src=src,
            automation_params=automation_params,
            silent_compilation_warnings=silent_compilation_warnings,
        )
        if queue is not None:
            self.ctx = queue.context
            self.queue = queue
        else:
            self.ctx = ctx
            self.queue = None
        self.compile_kernel_source(kernel_name, build_kwargs)
        self.get_kernel()

    def compile_kernel_source(self, kernel_name, build_kwargs):
        self.build_kwargs = build_kwargs
        self.kernel_name = kernel_name
        with catch_warnings(action=("ignore" if self.silent_compilation_warnings else None)):  # pylint: disable=E1123
            self.program = Program(self.ctx, self.src).build(**self.build_kwargs)

    def get_kernel(self):
        self.kernel = None
        for kern in self.program.all_kernels():
            if kern.function_name == self.kernel_name:
                self.kernel = kern
        if self.kernel is None:
            raise ValueError(
                "Could not find a kernel with function name '%s'. Available are: %s"
                % (self.kernel_name, self.program.kernel_names)
            )

    # overwrite parent method
    def guess_block_size(self, shape):
        device = self.ctx.devices[0]
        wg_max = device.max_work_group_size
        wg_multiple = self.kernel.get_work_group_info(kernel_work_group_info.PREFERRED_WORK_GROUP_SIZE_MULTIPLE, device)
        ndim = len(shape)
        # Try to have workgroup relatively well-balanced in all dimensions,
        # with more work items in x > y > z
        if ndim == 1:
            wg = (wg_max, 1, 1)
        else:
            w = (wg_max // wg_multiple, wg_multiple)
            wg = w if w[0] > w[1] else w[::-1]
            wg = wg + (1,)
            if ndim == 3:
                (wg[0] // 2, wg[1] // 4, 8)
        return wg

    def get_block_grid(self, *args, **kwargs):
        local_size = None
        global_size = block = None
        # COMPAT.
        block = kwargs.pop("block", None)
        if block is not None:
            deprecation_warning("Please use 'local_size' instead of 'block'")
        grid = kwargs.pop("grid", None)
        if grid is not None:
            deprecation_warning("Please use 'global_size' instead of 'grid'")
            global_size = tuple(g * b for g, b in zip(grid, block))
        #

        global_size = kwargs.pop("global_size", global_size)
        local_size = kwargs.pop("local_size", block)

        if global_size is None:
            raise ValueError("Need to define global_size for kernel '%s'" % self.kernel_name)

        if len(global_size) == 2 and local_size is not None and len(local_size) == 3:
            local_size = local_size[:-1]  # TODO check that last dim is 1

        self.last_block_size = local_size
        self.last_grid_size = global_size
        return local_size, global_size

    def follow_device_arr(self, args):
        args = list(args)
        for i, arg in enumerate(args):
            if isinstance(arg, parray.Array):
                args[i] = arg.data
        return tuple(args)

    def call(self, *args, **kwargs):
        if not isinstance(args[0], CommandQueue):
            queue = self.queue
            if queue is None:
                raise ValueError(
                    "First argument must be a pyopencl queue - otherwise provide OpenCLKernel(..., queue=queue)"
                )
        else:
            queue = args[0]
            args = args[1:]
        global_size, local_size, args, kwargs = self._prepare_call(*args, **kwargs)

        kwargs.pop("global_size", None)
        kwargs.pop("local_size", None)
        kwargs.pop("grid", None)
        kwargs.pop("block", None)

        return self.kernel(queue, global_size, local_size, *args, **kwargs)

    __call__ = call



import math
import numpy as np

from ..utils import get_cuda_srcfile
from ..cuda.processing import __has_pycuda__

if __has_pycuda__:
    from ..cuda.kernel import CudaKernel
    from .sinogram_cuda import CudaSinoMult

from .fbp import CudaBackprojector


try:
    import pycuda.driver as cuda
    from pycuda import gpuarray as garray

    __have_hbp__ = True
except:
    __have_hbp__ = False


def buildConebeamGeometry(
    anglesRad, rotAxisProjectionFromLeftPixelUnits, sourceSampleDistanceVoxelUnits, opticalAxisFromLeftPixelUnits=None
):
    """Generate fanbeam/conebeam projection matrices (as required by the backprojector) based on geometry parameters"""
    if opticalAxisFromLeftPixelUnits is None:
        if hasattr(rotAxisProjectionFromLeftPixelUnits, "__iter__"):
            opticalAxisFromLeftPixelUnits = rotAxisProjectionFromLeftPixelUnits[0]
        else:
            opticalAxisFromLeftPixelUnits = rotAxisProjectionFromLeftPixelUnits

    t = opticalAxisFromLeftPixelUnits
    d = sourceSampleDistanceVoxelUnits

    if hasattr(rotAxisProjectionFromLeftPixelUnits, "__iter__"):
        P_list = [
            np.array([[0, -t / d, 1, a], [1, 0, 0, 0], [0, -1 / d, 0, 1]], dtype=np.float64)  # pylint: disable=E1130
            for a in rotAxisProjectionFromLeftPixelUnits
        ]
    else:
        a = rotAxisProjectionFromLeftPixelUnits
        P_list = [
            np.array([[0, -t / d, 1, a], [1, 0, 0, 0], [0, -1 / d, 0, 1]], dtype=np.float64)  # pylint: disable=E1130
        ] * len(anglesRad)

    R = lambda w: np.array(
        [[1, 0, 0, 0], [0, np.cos(w), np.sin(w), 0], [0, -np.sin(w), np.cos(w), 0], [0, 0, 0, 1]], dtype=np.float64
    )
    return np.array([P @ R(-w) for P, w in zip(P_list, anglesRad)])


class HierarchicalBackprojector(CudaBackprojector):
    kernel_filename = "hierarchical_backproj.cu"

    def _init_geometry(self, sino_shape, slice_shape, angles, rot_center, halftomo, slice_roi):
        super()._init_geometry(sino_shape, slice_shape, angles, rot_center, halftomo, slice_roi)
        # pylint: disable=E1130 # -angles because different convention for the rotation direction
        self.angles = -self.angles

        # to do the reconstruction in reduction_steps steps
        self.reduction_steps = self.extra_options.get("hbp_reduction_steps", 2)
        reduction_factor = int(math.ceil((sino_shape[-2]) ** (1 / self.reduction_steps)))

        # TODO customize
        axis_source_meters = 1.0e9
        voxel_size_microns = 1.0
        #

        axis_cor = self.extra_options.get("axis_correction", None)
        if axis_cor is None:
            axis_cor = 0
        bpgeometry = buildConebeamGeometry(
            self.angles, self.rot_center + axis_cor, 1.0e6 * axis_source_meters / voxel_size_microns
        )
        self.setup_hbp(bpgeometry, reductionFactor=reduction_factor, legs=self.extra_options.get("hbp_legs", 4))

    def setup_hbp(
        self,
        bpgeometry,
        reductionFactor=20,
        grid_wh_factors=(1, 1),
        fac=1,
        legs=4,
    ):

        # This implementation seems not to use textures
        self._use_textures = False

        # for the non texture implementation, this big number will discard texture limitations
        large_factor_for_non_texture_memory_access = 2**10
        # TODO: read limits from device info.
        self.GPU_MAX_GRIDSIZE = 2**15 * large_factor_for_non_texture_memory_access
        self.GPU_MAX_GRIDS = 2**11 * large_factor_for_non_texture_memory_access

        if self.sino_shape[0] != len(bpgeometry):
            raise ValueError("self.sino_shape[0] != len(bpgeometry)")
        if self.sino_shape[0] != len(self.angles):
            raise ValueError("self.sino_shape[0] != len(self.angles)")

        if self.sino_shape[1] > self.GPU_MAX_GRIDSIZE:
            raise ValueError(f"self.sino_shape[1] > {self.GPU_MAX_GRIDSIZE} not supported by GPU")
        if self.sino_shape[0] > self.GPU_MAX_GRIDSIZE:
            raise ValueError(f"self.sino_shape[0] > {self.GPU_MAX_GRIDSIZE} currently not supported")

        self.reductionFactor = reductionFactor
        self.legs = legs

        self.bpsetupsH = bpgeometry.astype(np.float32)
        # self.bpsetupsD = cuda.mem_alloc(self.bpsetupsH.nbytes)
        # cuda.memcpy_htod(self.bpsetupsD, self.bpsetupsH)
        self.bpsetupsD = self._processing.to_device("bpsetupsD", self.bpsetupsH)

        # if allocate_cuda_sinogram:
        #     self.sinogramD = cuda.mem_alloc(self.sino_shape[0] * self.sino_shape[1] * self.float_size)
        # else:
        #     self.sinogramD = None
        self.sinogramD = None

        self.whf = grid_wh_factors
        if self.sino_shape[1] * 2 * self.whf[0] * fac > self.GPU_MAX_GRIDSIZE:
            print(f"WARNING: gridsampling limited to {self.GPU_MAX_GRIDSIZE}")
            self.whf[0] = self.GPU_MAX_GRIDSIZE / (self.sino_shape[1] * 2 * fac)

        ###############################################
        ########## create intermediate grids ##########
        ###############################################

        self.reductionFactors = []
        self.grids = []  # shapes
        self.gridTransforms = []  # grid-to-world
        self.gridInvTransforms = []  # world-to-grid
        self.gridTransformsH = []  # host buffer
        self.gridTransformsD = []  # device buffer

        ### first level grid: will receive backprojections #
        ####################################################

        N = self.slice_shape[1] * fac

        angularRange = abs(self.angles.ptp()) / self.sino_shape[0] * reductionFactor

        ngrids = int(math.ceil(self.sino_shape[0] / reductionFactor))

        grid_width = int(
            np.rint(2 * N * self.whf[0])
        )  # double sampling to account/compensate for diamond shaped grid of ray-intersections
        grid_height = int(
            math.ceil(angularRange * N * self.whf[1])
        )  # small-angle approximation, generates as much "lines" as needed to account for all intersection levels

        m = (len(self.angles) // reductionFactor) * reductionFactor
        # TODO: improve angle calculation for more general cases
        tmpangles = np.angle(
            np.average(np.exp(1.0j * self.angles[:m].reshape(m // reductionFactor, reductionFactor)), axis=1)
        )

        tmpangles = np.concatenate((tmpangles, (np.angle(np.average(np.exp(1.0j * self.angles[m:]))),)))[:ngrids]
        gridAinvT = self._getAinvT(N, grid_height, grid_width)
        setupRs = self._getRotationMatrices(tmpangles)

        pad = int(math.ceil(ngrids / legs) * legs - ngrids)  # add nan-padding for inline-signaling of unused grids
        self.gridTransforms += [
            np.array(
                [(R @ gridAinvT) for R in setupRs] + [np.ones((3, 3), np.float32) * math.nan] * pad, dtype=np.float32
            )
        ]
        self.gridInvTransforms += [np.array([np.linalg.inv(t) for t in self.gridTransforms[-1]], dtype=np.float32)]
        self.grids += [(grid_height, grid_width, int(math.ceil(ngrids / legs)))]
        self.reductionFactors += [reductionFactor]

        ### intermediate level grids: accumulation grids ###
        ####################################################

        # Actual iteration count typically within 1-5. Cf. break condition
        for i in range(100):
            # for a reasonable (with regard to memory requirement) grid-aspect ratio in the intermediate levels,
            # the covered angular range per grid should not exceed 28.6°, i.e.,
            # fewer than 7 (6.3) or 13 (12.6)  grids for a 180° / 360° scan is not reasonable
            if int(math.ceil(ngrids / reductionFactor)) < 20:
                break
            angularRange *= reductionFactor
            ngrids = int(math.ceil(ngrids / reductionFactor))

            grid_height = int(
                math.ceil(angularRange * N * self.whf[1])
            )  # implicit small angle approximation, whose validity is
            # asserted by the preceding "break"
            gridAinvT = self._getAinvT(N, grid_height, grid_width)

            prevAngles = tmpangles
            m = (len(prevAngles) // reductionFactor) * reductionFactor
            # TODO: improve angle calculation for more general cases
            tmpangles = np.angle(
                np.average(np.exp(1.0j * prevAngles[:m].reshape(m // reductionFactor, reductionFactor)), axis=1)
            )
            tmpangles = np.concatenate((tmpangles, (np.angle(np.average(np.exp(1.0j * prevAngles[m:]))),)))[:ngrids]
            setupRsRed = self._getRotationMatrices(tmpangles)

            pad = int(math.ceil(ngrids / legs) * legs - ngrids)
            self.gridTransforms += [
                np.array(
                    [(R @ gridAinvT) for R in setupRsRed] + [np.ones((3, 3), np.float32) * math.nan] * pad,
                    dtype=np.float32,
                )
            ]
            self.gridInvTransforms += [np.array([np.linalg.inv(t) for t in self.gridTransforms[-1]], dtype=np.float32)]
            self.grids += [(grid_height, grid_width, int(math.ceil(ngrids / legs)))]
            self.reductionFactors += [reductionFactor]

        ##### final accumulation grid #################
        ###############################################

        reductionFactor = ngrids
        ngrids = 1
        grid_size = self.slice_shape[1]
        grid_width = grid_size
        grid_height = grid_size

        # gridAinvT    = self._getAinvT(N, grid_height, grid_width)
        gridAinvT = self._getAinvT(N, grid_height, grid_width, 1 / fac)

        self.gridTransforms += [
            np.array([gridAinvT] * legs, dtype=np.float32)
        ]  # inflate transform list for convenience in reconstruction loop
        self.gridInvTransforms += [np.array([np.linalg.inv(t) for t in self.gridTransforms[-1]], dtype=np.float32)]
        self.grids += [(grid_height, grid_width, ngrids)]
        self.reductionFactors += [reductionFactor]

        #### accumulation grids #####
        self.gridTransformsD = []
        self.gridInvTransformsD = []
        self.gridsD = []

        max_grid_size = get_max_grid_size(self.grids)

        for i in range(len(self.grids)):
            gridTransformH = np.array(self.gridTransforms[i][:, :2, :3], dtype=np.float32, order="C").copy()
            gridInvTransformH = np.array(self.gridInvTransforms[i][:, :2, :3], dtype=np.float32, order="C").copy()
            self.gridTransformsD.append(self._processing.to_device("gridTransformsD%d " % i, gridTransformH.ravel()))
            self.gridInvTransformsD.append(
                self._processing.to_device("gridInvTransformsD%d" % i, gridInvTransformH.ravel())
            )

            if legs == 1 or i + 1 != (len(self.grids)):
                if i < 2:
                    self.gridsD.append(self._processing.allocate_array("gridsD%d" % i, max_grid_size))
                else:
                    self.gridsD.append(self.gridsD[i % 2])
            else:
                self.gridsD.append(self._processing.allocate_array("gridsD%d" % i, get_max_grid_size(self.grids[-1:])))

        self.imageBufferShape = (grid_size, grid_size)
        self.imageBufferD = self._processing.allocate_array(
            "imageBufferD", self.imageBufferShape[0] * self.imageBufferShape[1]
        )
        self.imageBufferH = np.zeros(self.imageBufferShape, dtype=np.float32)

    def _getAinvT(self, finalGridWidthAndHeight, currentGridHeight, currentGridWidth, scale=1):
        N = finalGridWidthAndHeight
        grid_height = currentGridHeight
        grid_width = currentGridWidth

        # shifts a texture coordinate from corner origin to center origin
        T = np.array(((1, 0, -0.5 * (grid_height - 1)), (0, 1, -0.5 * (grid_width - 1)), (0, 0, 1)), dtype=np.float32)
        # scales texture coordinates (of subsampled grid) into the unit/cooridnate system of a fully sampled grid
        Ainv = np.array(
            (((N - 1) / (grid_height - 1) * scale, 0, 0), (0, (N - 1) / (grid_width - 1) * scale, 0), (0, 0, 1)),
            dtype=np.float32,
        )
        return Ainv @ T

    def _getRotationMatrices(self, angles):
        return [
            np.array(((np.cos(a), np.sin(a), 0), (-np.sin(a), np.cos(a), 0), (0, 0, 1)), dtype=np.float32)
            for a in angles
        ]

    def _compile_kernels(self):
        # pylint: disable=E0606
        self.backprojector = CudaKernel(
            "backprojector",
            filename=get_cuda_srcfile(self.kernel_filename),
            signature="PPiiiiPiifPi",
        )
        self.aggregator = CudaKernel(
            "aggregator", filename=get_cuda_srcfile(self.kernel_filename), signature="iPPiiiiPiiiP"
        )
        self.clip_outer_circle_kernel = CudaKernel(
            "clip_outer_circle", filename=get_cuda_srcfile(self.kernel_filename), signature="Pii"
        )
        # Duplicate of fbp.py ...
        if self.halftomo and self.rot_center < self.dwidth:
            self.sino_mult = CudaSinoMult(self.sino_shape, self.rot_center, ctx=self._processing.ctx)
        #

    def _set_sino(self, sino, do_checks=True):
        if do_checks and not (sino.flags.c_contiguous):
            raise ValueError("Expected C-Contiguous array")
        else:
            self._d_sino = self._processing.allocate_array("_d_sino", self.sino_shape)
            if id(self._d_sino) == id(sino):
                return
            self._d_sino[:] = sino[:]

    def backproj(self, sino, output=None, do_checks=True, reference=False):
        if self.halftomo and self.rot_center < self.dwidth:
            self.sino_mult.prepare_sino(sino)
        self._set_sino(sino)
        lws = (64, 4, 4)

        if reference:
            gws = getGridSize(self.grids[-1], lws)
            (grid_height, grid_width, ngrids) = self.grids[-1]

            self.backprojector(
                self.bpsetupsD,
                self.gridTransformsD[-1].gpudata,
                np.int32(self.sino_shape[0]),
                np.int32(grid_width),
                np.int32(grid_height),
                np.int32(ngrids),
                self.gridsD[-1],
                np.int32(self.sino_shape[1]),
                np.int32(self.sino_shape[0]),
                np.float32(self._backproj_scale_factor),
                self._d_sino,
                np.int32(0),  # offset
                block=lws,
                grid=gws,
            )

        else:
            for leg in list(range(0, self.legs)):
                gridOffset = leg * self.grids[0][2]
                projOffset = gridOffset * self.reductionFactors[0]
                gws = getGridSize(self.grids[0], lws)
                (grid_height, grid_width, ngrids) = self.grids[0]

                self.backprojector(
                    self.bpsetupsD,
                    self.gridTransformsD[0][6 * gridOffset :],
                    np.int32(self.reductionFactors[0]),
                    np.int32(grid_width),
                    np.int32(grid_height),
                    np.int32(ngrids),
                    self.gridsD[0],
                    np.int32(self.sino_shape[1]),
                    np.int32(self.sino_shape[0]),
                    np.float32(self._backproj_scale_factor),
                    self._d_sino,
                    np.int32(projOffset),
                    block=lws,
                    grid=gws,
                )

                for i in range(1, len(self.grids)):
                    if self.grids[i][2] >= 8:
                        lws = (16, 16, 4)
                    else:
                        lws = (32, 32, 1)

                    gws = getGridSize(self.grids[i], lws)

                    (new_grid_height, new_grid_width, new_ngrids) = self.grids[i]
                    (prev_grid_height, prev_grid_width, prev_ngrids) = self.grids[i - 1]

                    gridOffset = leg * self.grids[i][2]
                    prevGridOffset = leg * self.grids[i - 1][2]

                    self.aggregator(
                        np.int32((i + 1 == len(self.grids)) and (leg > 0)),
                        self.gridTransformsD[i][6 * gridOffset :],
                        self.gridInvTransformsD[i - 1][6 * prevGridOffset :],
                        np.int32(self.reductionFactors[i]),
                        np.int32(new_grid_width),
                        np.int32(new_grid_height),
                        np.int32(new_ngrids),
                        self.gridsD[i],
                        np.int32(prev_grid_width),
                        np.int32(prev_grid_height),
                        np.int32(prev_ngrids),
                        self.gridsD[i - 1],
                        block=lws,
                        grid=gws,
                    )

        if self.extra_options.get("clip_outer_circle", False):
            lws = (16, 16, 1)
            ny, nx = self.slice_shape
            gws = getGridSize((nx, ny, 1), lws)
            self.clip_outer_circle_kernel(self.gridsD[-1], np.int32(ny), np.int32(nx), block=lws, grid=gws)

        # FIXME pycuda fails to do a discontiguous memcpy for more than 2^31 bytes
        if self.gridsD[-1].nbytes > 2**31:
            r1d = self.gridsD[-1].get()
            r2d = np.ascontiguousarray(r1d.reshape(self.slice_shape))
            if output is not None:
                output[:] = r2d[:]
                return output
            else:
                return r2d
        # --------

        else:
            return self.gridsD[-1].reshape(self.slice_shape).get(ary=output)


def get_max_grid_size(grids):
    size_max = 0
    for dims in grids:
        size = 1
        for d in dims:
            size = size * d
        if size > size_max:
            size_max = size
    return size_max


def getGridSize(minimum, local):
    m, l = np.array(minimum), np.array(local)
    new = (m // l) * l
    new[new < m] += l[new < m]
    return tuple(map(int, new // l))



import numpy as np
import pytest
from scipy.interpolate import interp1d
from nabu.testutils import generate_tests_scenarios, get_data
from nabu.utils import get_cuda_srcfile, updiv
from nabu.cuda.utils import __has_pycuda__, get_cuda_context

if __has_pycuda__:
    import pycuda.gpuarray as garray
    from nabu.cuda.kernel import CudaKernel


img0 = get_data("brain_phantom.npz")["data"]

scenarios = generate_tests_scenarios(
    {
        "image": [img0, img0[:, :511], img0[:511, :]],
        "x_bounds": [(180, 360), (0, 180), (50, 50 + 180)],
        "x_to_x_new": [0.1, -0.2, 0.3],
    }
)


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    cls.tol = 1e-4
    if __has_pycuda__:
        cls.ctx = get_cuda_context(cleanup_at_exit=False)
    yield
    if __has_pycuda__:
        cls.ctx.pop()


@pytest.mark.usefixtures("bootstrap")
class TestInterpolation:
    def _get_reference_interpolation(self, img, x, x_new):
        interpolator = interp1d(x, img, kind="linear", axis=0, fill_value="extrapolate", copy=True)
        ref = interpolator(x_new)
        return ref

    def _compare(self, res, img, x, x_new):
        ref = self._get_reference_interpolation(img, x, x_new)
        mae = np.max(np.abs(res - ref))
        return mae

    # parametrize on a class method will use the same class, and launch this
    # method with different scenarios.
    @pytest.mark.skipif(not (__has_pycuda__), reason="need pycuda for this test")
    @pytest.mark.parametrize("config", scenarios)
    def test_cuda_interpolation(self, config):
        img = config["image"]
        Ny, Nx = img.shape

        xmin, xmax = config["x_bounds"]
        x = np.linspace(xmin, xmax, num=img.shape[0], endpoint=False, dtype="f")
        x_new = x + config["x_to_x_new"]

        d_img = garray.to_gpu(img)
        d_out = garray.zeros_like(d_img)
        d_x = garray.to_gpu(x)
        d_x_new = garray.to_gpu(x_new)

        cuda_interpolator = CudaKernel(
            "linear_interp_vertical", get_cuda_srcfile("interpolation.cu"), signature="PPiiPP"
        )
        cuda_interpolator(d_img, d_out, Nx, Ny, d_x, d_x_new, grid=(updiv(Nx, 16), updiv(Ny, 16), 1), block=(16, 16, 1))
        err = self._compare(d_out.get(), img, x, x_new)
        err_msg = str("Max error is too high for this configuration: %s" % str(config))
        assert err < self.tol, err_msg



import logging
import numpy
import os
import h5py
from typing import Union
from nabu.stitching.config import PostProcessedSingleAxisStitchingConfiguration
from nabu.stitching.alignment import AlignmentAxis1
from nabu.stitching.alignment import PaddedRawData
from math import ceil
from tomoscan.io import HDF5File
from tomoscan.esrf.scan.utils import cwd_context
from tomoscan.esrf import NXtomoScan
from tomoscan.series import Series
from tomoscan.volumebase import VolumeBase
from tomoscan.esrf.volume import HDF5Volume
from typing import Iterable
from contextlib import AbstractContextManager
from pyunitsystem.metricsystem import MetricSystem
from nabu.stitching.config import (
    PostProcessedSingleAxisStitchingConfiguration,
    KEY_IMG_REG_METHOD,
)
from nabu.stitching.utils.utils import find_volumes_relative_shifts
from nabu.io.utils import DatasetReader
from .single_axis import SingleAxisStitcher

_logger = logging.getLogger(__name__)


class FlippingValueError(ValueError):
    pass


class PostProcessingStitching(SingleAxisStitcher):
    """
    Loader to be used when load data during post-processing stitching (on recosntructed volume). Output is expected to be an NXtomo
    """

    def __init__(self, configuration, progress=None) -> None:
        if not isinstance(configuration, PostProcessedSingleAxisStitchingConfiguration):
            raise TypeError(
                f"configuration is expected to be an instance of {PostProcessedSingleAxisStitchingConfiguration}. Get {type(configuration)} instead"
            )
        self._input_volumes = configuration.input_volumes
        self.__output_data_type = None

        self._series = Series("series", iterable=self._input_volumes, use_identifiers=False)

        super().__init__(configuration, progress=progress)

    @property
    def stitching_axis_in_frame_space(self):
        if self.axis == 0:
            return 0
        elif self.axis in (1, 2):
            raise NotImplementedError(f"post-processing stitching along axis {self.axis} is not handled.")
        else:
            raise NotImplementedError(f"stitching axis must be in (0, 1, 2). Get {self.axis}")

    def settle_flips(self):
        super().settle_flips()
        if not self.configuration.duplicate_data:
            if len(numpy.unique(self.configuration.flip_lr)) > 1:
                raise FlippingValueError(
                    "Stitching without data duplication cannot handle volume with different flip. Please run the stitching with data duplication"
                )
            if True in self.configuration.flip_ud:
                raise FlippingValueError(
                    "Stitching without data duplication cannot handle with up / down flips. Please run the stitching with data duplication"
                )

    def order_input_tomo_objects(self):

        def get_min_bound(volume):
            try:
                bb = volume.get_bounding_box(axis=self.axis)
            except ValueError:  #  if missing information
                bb = None
            if bb is not None:
                return bb.min
            else:
                # if can't find bounding box (missing metadata to the volume
                # try to get it from the scan
                metadata = volume.metadata or volume.load_metadata()
                scan_location = metadata.get("nabu_config", {}).get("dataset", {}).get("location", None)
                scan_entry = metadata.get("nabu_config", {}).get("dataset", {}).get("hdf5_entry", None)
                if scan_location is not None:
                    # this work around (until most volume have position metadata) works only for Hdf5volume
                    with cwd_context(os.path.dirname(volume.file_path)):
                        o_scan = NXtomoScan(scan_location, scan_entry)
                        bb_acqui = o_scan.get_bounding_box(axis=None)
                        # for next step volume position will be required.
                        # if you can find it set it directly
                        volume.position = (numpy.array(bb_acqui.max) - numpy.array(bb_acqui.min)) / 2.0 + numpy.array(
                            bb_acqui.min
                        )
                        # for now translation are stored in pixel size ref instead of real_pixel_size
                        volume.pixel_size = o_scan.x_real_pixel_size
                        if bb_acqui is not None:
                            return bb_acqui.min[0]
                raise ValueError("Unable to find volume position. Unable to deduce z position")

        try:
            # order volumes from higher z to lower z
            # if axis 0 position is provided then use directly it
            if self.configuration.axis_0_pos_px is not None and len(self.configuration.axis_0_pos_px) > 0:
                order = numpy.argsort(self.configuration.axis_0_pos_px)
                sorted_series = Series(
                    self.series.name,
                    numpy.take_along_axis(numpy.array(self.series[:]), order, axis=0)[::-1],
                    use_identifiers=False,
                )
            else:
                # else use bounding box
                sorted_series = Series(
                    self.series.name,
                    sorted(self.series[:], key=get_min_bound, reverse=True),
                    use_identifiers=False,
                )
        except ValueError:
            _logger.warning(
                "Unable to find volume positions in metadata. Expect the volume to be ordered already (decreasing along axis 0.)"
            )
        else:
            if sorted_series == self.series:
                pass
            elif sorted_series != self.series:
                if sorted_series[:] != self.series[::-1]:
                    raise ValueError(
                        "Unable to get comprehensive input. ordering along axis 0 is not respected (decreasing)."
                    )
                else:
                    _logger.warning(
                        f"decreasing order haven't been respected. Need to reorder {self.serie_label} ({[str(scan) for scan in sorted_series[:]]}). Will also reorder positions"
                    )
                    if self.configuration.axis_0_pos_mm is not None:
                        self.configuration.axis_0_pos_mm = self.configuration.axis_0_pos_mm[::-1]
                    if self.configuration.axis_0_pos_px is not None:
                        self.configuration.axis_0_pos_px = self.configuration.axis_0_pos_px[::-1]
                    if self.configuration.axis_1_pos_mm is not None:
                        self.configuration.axis_1_pos_mm = self.configuration.axis_1_pos_mm[::-1]
                    if self.configuration.axis_1_pos_px is not None:
                        self.configuration.axis_1_pos_px = self.configuration.axis_1_pos_px[::-1]
                    if self.configuration.axis_2_pos_mm is not None:
                        self.configuration.axis_2_pos_mm = self.configuration.axis_2_pos_mm[::-1]
                    if self.configuration.axis_2_pos_px is not None:
                        self.configuration.axis_2_pos_px = self.configuration.axis_2_pos_px[::-1]
                    if not numpy.isscalar(self._configuration.flip_ud):
                        self._configuration.flip_ud = self._configuration.flip_ud[::-1]
                    if not numpy.isscalar(self._configuration.flip_lr):
                        self._configuration.flip_ud = self._configuration.flip_lr[::-1]

                    self._series = sorted_series

    def check_inputs(self):
        """
        insure input data is coherent
        """
        # check input volume
        if self.configuration.output_volume is None:
            raise ValueError("input volume should be provided")

        n_volumes = len(self.series)
        if n_volumes == 0:
            raise ValueError("no scan to stich together")

        if not isinstance(self.configuration.output_volume, VolumeBase):
            raise TypeError(f"make sure we return a volume identifier not {(type(self.configuration.output_volume))}")

        # check axis 0 position
        if isinstance(self.configuration.axis_0_pos_px, Iterable) and len(self.configuration.axis_0_pos_px) != (
            n_volumes
        ):
            raise ValueError(f"expect {n_volumes} overlap defined. Get {len(self.configuration.axis_0_pos_px)}")
        if isinstance(self.configuration.axis_0_pos_mm, Iterable) and len(self.configuration.axis_0_pos_mm) != (
            n_volumes
        ):
            raise ValueError(f"expect {n_volumes} overlap defined. Get {len(self.configuration.axis_0_pos_mm)}")

        # check axis 1 position
        if isinstance(self.configuration.axis_1_pos_px, Iterable) and len(self.configuration.axis_1_pos_px) != (
            n_volumes
        ):
            raise ValueError(f"expect {n_volumes} overlap defined. Get {len(self.configuration.axis_1_pos_px)}")
        if isinstance(self.configuration.axis_1_pos_mm, Iterable) and len(self.configuration.axis_1_pos_mm) != (
            n_volumes
        ):
            raise ValueError(f"expect {n_volumes} overlap defined. Get {len(self.configuration.axis_1_pos_mm)}")

        # check axis 2 position
        if isinstance(self.configuration.axis_1_pos_px, Iterable) and len(self.configuration.axis_1_pos_px) != (
            n_volumes
        ):
            raise ValueError(f"expect {n_volumes} overlap defined. Get {len(self.configuration.axis_1_pos_px)}")
        if isinstance(self.configuration.axis_2_pos_mm, Iterable) and len(self.configuration.axis_2_pos_mm) != (
            n_volumes
        ):
            raise ValueError(f"expect {n_volumes} overlap defined. Get {len(self.configuration.axis_2_pos_mm)}")

        self._reading_orders = []
        # the first scan will define the expected reading orderd, and expected flip.
        # if all scan are flipped then we will keep it this way
        self._reading_orders.append(1)

    @staticmethod
    def _get_bunch_of_data(
        bunch_start: int,
        bunch_end: int,
        step: int,
        volumes: tuple,
        flip_lr_arr: bool,
        flip_ud_arr: bool,
    ):
        """
        goal is to load contiguous frames as much as possible...
        return for each volume the bunch of slice along axis 1
        warning: they can have different shapes
        """

        def get_sub_volume(volume, flip_lr, flip_ud):
            sub_volume = volume[:, bunch_start:bunch_end:step, :]
            if flip_lr:
                sub_volume = numpy.fliplr(sub_volume)
            if flip_ud:
                sub_volume = numpy.flipud(sub_volume)
            return sub_volume

        sub_volumes = [
            get_sub_volume(volume, flip_lr, flip_ud)
            for volume, flip_lr, flip_ud in zip(volumes, flip_lr_arr, flip_ud_arr)
        ]
        # generator on it self: we want to iterate over the y axis
        n_slices_in_bunch = ceil((bunch_end - bunch_start) / step)
        assert isinstance(n_slices_in_bunch, int)
        for i in range(n_slices_in_bunch):
            yield [sub_volume[:, i, :] for sub_volume in sub_volumes]

    def compute_estimated_shifts(self):
        axis_0_pos_px = self.configuration.axis_0_pos_px
        self._axis_0_rel_ini_shifts = []
        # compute overlap along axis 0
        for upper_volume, lower_volume, upper_volume_axis_0_pos, lower_volume_axis_0_pos in zip(
            self.series[:-1], self.series[1:], axis_0_pos_px[:-1], axis_0_pos_px[1:]
        ):
            upper_volume_low_pos = upper_volume_axis_0_pos - upper_volume.get_volume_shape()[0] / 2
            lower_volume_high_pos = lower_volume_axis_0_pos + lower_volume.get_volume_shape()[0] / 2
            self._axis_0_rel_ini_shifts.append(
                int(lower_volume_high_pos - upper_volume_low_pos)  # overlap are expected to be int for now
            )
        self._axis_1_rel_ini_shifts = self.from_abs_pos_to_rel_pos(self.configuration.axis_1_pos_px)
        self._axis_2_rel_ini_shifts = [0.0] * (len(self.series) - 1)

    def _compute_positions_as_px(self):
        """compute if necessary position other axis 0 from volume metadata"""

        def get_position_as_px_on_axis(axis, pos_as_px, pos_as_mm):
            if pos_as_px is not None:
                if pos_as_mm is not None:
                    raise ValueError(
                        f"position of axis {axis} is provided twice: as mm and as px. Please provide one only ({pos_as_mm} vs {pos_as_px})"
                    )
                else:
                    return pos_as_px

            elif pos_as_mm is not None:
                # deduce from position given in configuration and pixel size
                axis_N_pos_px = []
                for volume, pos_in_mm in zip(self.series, pos_as_mm):
                    voxel_size_m = self.configuration.voxel_size or volume.voxel_size
                    axis_N_pos_px.append((pos_in_mm / MetricSystem.MILLIMETER.value) / voxel_size_m[0])
                return axis_N_pos_px
            else:
                # deduce from motor position and pixel size
                axis_N_pos_px = []
                base_position_m = self.series[0].get_bounding_box(axis=axis).min
                for volume in self.series:
                    voxel_size_m = self.configuration.voxel_size or volume.voxel_size
                    volume_axis_bb = volume.get_bounding_box(axis=axis)
                    axis_N_mean_pos_m = (volume_axis_bb.max - volume_axis_bb.min) / 2 + volume_axis_bb.min
                    axis_N_mean_rel_pos_m = axis_N_mean_pos_m - base_position_m
                    axis_N_pos_px.append(int(axis_N_mean_rel_pos_m / voxel_size_m[0]))
                return axis_N_pos_px

        self.configuration.axis_0_pos_px = get_position_as_px_on_axis(
            axis=0,
            pos_as_px=self.configuration.axis_0_pos_px,
            pos_as_mm=self.configuration.axis_0_pos_mm,
        )
        self.configuration.axis_0_pos_mm = None

        self.configuration.axis_1_pos_px = get_position_as_px_on_axis(
            axis=1,
            pos_as_px=self.configuration.axis_1_pos_px,
            pos_as_mm=self.configuration.axis_1_pos_mm,
        )

        self.configuration.axis_2_pos_px = get_position_as_px_on_axis(
            axis=2,
            pos_as_px=self.configuration.axis_2_pos_px,
            pos_as_mm=self.configuration.axis_2_pos_mm,
        )
        self.configuration.axis_2_pos_mm = None

    def _compute_shifts(self):
        n_volumes = len(self.configuration.input_volumes)
        if n_volumes == 0:
            raise ValueError("no scan to stich provided")

        slice_for_shift = self.configuration.slice_for_cross_correlation or "middle"
        y_rel_shifts = self._axis_0_rel_ini_shifts
        x_rel_shifts = self._axis_1_rel_ini_shifts
        dim_axis_1 = max([volume.get_volume_shape()[1] for volume in self.series])

        final_rel_shifts = []
        for (
            upper_volume,
            lower_volume,
            x_rel_shift,
            y_rel_shift,
            flip_ud_upper,
            flip_ud_lower,
        ) in zip(
            self.series[:-1],
            self.series[1:],
            x_rel_shifts,
            y_rel_shifts,
            self.configuration.flip_ud[:-1],
            self.configuration.flip_ud[1:],
        ):
            x_cross_algo = self.configuration.axis_2_params.get(KEY_IMG_REG_METHOD, None)
            y_cross_algo = self.configuration.axis_0_params.get(KEY_IMG_REG_METHOD, None)

            # compute relative shift
            found_shift_y, found_shift_x = find_volumes_relative_shifts(
                upper_volume=upper_volume,
                lower_volume=lower_volume,
                dtype=self.get_output_data_type(),
                dim_axis_1=dim_axis_1,
                slice_for_shift=slice_for_shift,
                x_cross_correlation_function=x_cross_algo,
                y_cross_correlation_function=y_cross_algo,
                x_shifts_params=self.configuration.axis_2_params,
                y_shifts_params=self.configuration.axis_0_params,
                estimated_shifts=(y_rel_shift, x_rel_shift),
                flip_ud_lower_frame=flip_ud_lower,
                flip_ud_upper_frame=flip_ud_upper,
                alignment_axis_1=self.configuration.alignment_axis_1,
                alignment_axis_2=self.configuration.alignment_axis_2,
                overlap_axis=self.axis,
            )
            final_rel_shifts.append(
                (found_shift_y, found_shift_x),
            )

        # set back values. Now position should start at 0
        self._axis_0_rel_final_shifts = [final_shift[0] for final_shift in final_rel_shifts]
        self._axis_1_rel_final_shifts = [final_shift[1] for final_shift in final_rel_shifts]
        self._axis_2_rel_final_shifts = [0.0] * len(final_rel_shifts)
        _logger.info(f"axis 2 relative shifts (x in radio ref) to be used will be {self._axis_1_rel_final_shifts}")
        print(f"axis 2 relative shifts (x in radio ref) to be used will be {self._axis_1_rel_final_shifts}")
        _logger.info(f"axis 0 relative shifts (y in radio ref) y to be used will be {self._axis_0_rel_final_shifts}")
        print(f"axis 0 relative shifts (y in radio ref) y to be used will be {self._axis_0_rel_final_shifts}")

    def get_output_data_type(self):
        if self.__output_data_type is None:

            def find_output_data_type():
                first_vol = self._input_volumes[0]
                if first_vol.data is not None:
                    return first_vol.data.dtype
                elif isinstance(first_vol, HDF5Volume):
                    with DatasetReader(first_vol.data_url) as vol_dataset:
                        return vol_dataset.dtype
                else:
                    return first_vol.load_data(store=False).dtype

            self.__output_data_type = find_output_data_type()
        return self.__output_data_type

    def _create_stitched_volume(self, store_composition: bool):
        overlap_kernels = self._overlap_kernels
        self._slices_to_stitch, n_slices = self.configuration.settle_slices()

        # sync overwrite_results with volume overwrite parameter
        self.configuration.output_volume.overwrite = self.configuration.overwrite_results

        # init final volume
        final_volume = self.configuration.output_volume
        final_volume_shape = (
            int(
                numpy.asarray([volume.get_volume_shape()[0] for volume in self._input_volumes]).sum()
                - numpy.asarray([abs(overlap) for overlap in self._axis_0_rel_final_shifts]).sum(),
            ),
            n_slices,
            self._stitching_constant_length,
        )

        data_type = self.get_output_data_type()

        if self.progress:
            self.progress.total = final_volume_shape[1]

        y_index = 0
        if isinstance(self._slices_to_stitch, slice):
            step = self._slices_to_stitch.step or 1
        else:
            step = 1

        output_dataset_args = {
            "volume": final_volume,
            "volume_shape": final_volume_shape,
            "dtype": data_type,
            "dumper": self.dumper,
        }
        from .dumper.postprocessing import PostProcessingStitchingDumperNoDD

        # TODO: FIXME: for now not very elegant but in the case of avoiding data duplication
        # we need to provide the the information about the stitched part shape.
        # this should be move to the dumper in the future
        if isinstance(self.dumper, PostProcessingStitchingDumperNoDD):
            output_dataset_args["stitching_sources_arr_shapes"] = tuple(
                [(abs(overlap), n_slices, self._stitching_constant_length) for overlap in self._axis_0_rel_final_shifts]
            )

        with self.dumper.OutputDatasetContext(**output_dataset_args):
            # note: output_dataset is a HDF5 dataset if final volume is an HDF5 volume else is a numpy array
            with _RawDatasetsContext(
                self._input_volumes,
                alignment_axis_1=self.configuration.alignment_axis_1,
            ) as raw_datasets:
                # note: raw_datasets can be numpy arrays or HDF5 dataset (in the case of HDF5Volume)
                # to speed up we read by bunch of dataset. For numpy array this doesn't change anything
                # but for HDF5 dataset this can speed up a lot the processing (depending on HDF5 dataset chuncks)
                # note: we read trhough axis 1
                if isinstance(self.dumper, PostProcessingStitchingDumperNoDD):
                    self.dumper.raw_regions_hdf5_dataset = raw_datasets
                for bunch_start, bunch_end in PostProcessingStitching._data_bunch_iterator(
                    slices=self._slices_to_stitch, bunch_size=50
                ):
                    for data_frames in PostProcessingStitching._get_bunch_of_data(
                        bunch_start,
                        bunch_end,
                        step=step,
                        volumes=raw_datasets,
                        flip_lr_arr=self.configuration.flip_lr,
                        flip_ud_arr=self.configuration.flip_ud,
                    ):
                        if self.configuration.rescale_frames:
                            data_frames = self.rescale_frames(data_frames)
                        if self.configuration.normalization_by_sample.is_active():
                            data_frames = self.normalize_frame_by_sample(data_frames)

                        sf = PostProcessingStitching.stitch_frames(
                            frames=data_frames,
                            axis=self.axis,
                            output_dtype=data_type,
                            x_relative_shifts=self._axis_1_rel_final_shifts,
                            y_relative_shifts=self._axis_0_rel_final_shifts,
                            overlap_kernels=overlap_kernels,
                            dumper=self.dumper,
                            i_frame=y_index,
                            return_composition_cls=store_composition if y_index == 0 else False,
                            stitching_axis=self.axis,
                            check_inputs=y_index == 0,  # on process check on the first iteration
                        )
                        if y_index == 0 and store_composition:
                            _, self._frame_composition = sf

                        if self.progress is not None:
                            self.progress.update()
                        y_index += 1

    # alias to general API
    def _create_stitching(self, store_composition):
        self._create_stitched_volume(store_composition=store_composition)


class _RawDatasetsContext(AbstractContextManager):
    """
    return volume data for all input volume (target: used for volume stitching).
    If the volume is an HDF5Volume then the HDF5 dataset will be used (on disk)
    If the volume is of another type then it will be loaded in memory then used (more memory consuming)
    """

    def __init__(self, volumes: tuple, alignment_axis_1) -> None:
        super().__init__()
        for volume in volumes:
            if not isinstance(volume, VolumeBase):
                raise TypeError(
                    f"Volumes are expected to be an instance of {VolumeBase}. {type(volume)} provided instead"
                )

        self._volumes = volumes
        self.__file_handlers = []
        self._alignment_axis_1 = alignment_axis_1

    @property
    def alignment_axis_1(self):
        return self._alignment_axis_1

    def __enter__(self):
        # handle the specific case of HDF5. Goal: avoid getting the full stitched volume in memory
        datasets = []
        shapes = {volume.get_volume_shape()[1] for volume in self._volumes}
        axis_1_dim = max(shapes)
        axis_1_need_padding = len(shapes) > 1

        try:
            for volume in self._volumes:
                if volume.data is not None:
                    data = volume.data
                elif isinstance(volume, HDF5Volume):
                    file_handler = HDF5File(volume.data_url.file_path(), mode="r")
                    dataset = file_handler[volume.data_url.data_path()]
                    data = dataset
                    self.__file_handlers.append(file_handler)
                # for other file format: load the full dataset in memory
                else:
                    data = volume.load_data(store=False)
                    if data is None:
                        raise ValueError(f"No data found for volume {volume.get_identifier()}")
                if axis_1_need_padding:
                    data = self.add_padding(data=data, axis_1_dim=axis_1_dim, alignment=self.alignment_axis_1)
                datasets.append(data)
        except Exception as e:
            # if some errors happen during loading HDF5
            for file_handled in self.__file_handlers:
                file_handled.close()
            raise e

        return datasets

    def __exit__(self, exc_type, exc_value, traceback):
        success = True
        for file_handler in self.__file_handlers:
            success = success and file_handler.close()
        if exc_type is None:
            return success

    def add_padding(self, data: Union[h5py.Dataset, numpy.ndarray], axis_1_dim, alignment: AlignmentAxis1):
        alignment = AlignmentAxis1.from_value(alignment)
        if alignment is AlignmentAxis1.BACK:
            axis_1_pad_width = (axis_1_dim - data.shape[1], 0)
        elif alignment is AlignmentAxis1.CENTER:
            half_width = int((axis_1_dim - data.shape[1]) / 2)
            axis_1_pad_width = (half_width, axis_1_dim - data.shape[1] - half_width)
        elif alignment is AlignmentAxis1.FRONT:
            axis_1_pad_width = (0, axis_1_dim - data.shape[1])
        else:
            raise ValueError(f"alignment {alignment} is not handled")

        return PaddedRawData(
            data=data,
            axis_1_pad_width=axis_1_pad_width,
        )



import os.path as path
from math import exp
import tempfile
import numpy as np
import pytest
from silx.io.url import DataUrl
from tomoscan.esrf.mock import MockNXtomo
from nabu.io.reader import HDF5Reader
from nabu.preproc.double_flatfield import DoubleFlatField
from nabu.cuda.utils import __has_pycuda__, get_cuda_context

if __has_pycuda__:
    import pycuda.gpuarray as garray
    from nabu.preproc.double_flatfield_cuda import CudaDoubleFlatField, __has_pycuda__


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    cls.tmpdir = tempfile.TemporaryDirectory()
    dname = cls.tmpdir.name
    cls.dname = dname
    radios = MockNXtomo(
        path.join(dname, "tmp"),
        10,
        n_ini_proj=10,
        dim=100,
        n_refs=1,
        scene="increasing value",
    ).scan.projections
    reader = HDF5Reader()
    cls.radios = []
    Rkeys = list(radios.keys())
    for k in Rkeys:
        dataurl = radios[k]
        data = reader.get_data(dataurl)
        cls.radios.append(data)
    cls.radios = np.array(cls.radios)
    cls.ff_dump_url = DataUrl(
        file_path=path.join(cls.dname, "dff.h5"), data_path="/entry/double_flatfield/results/data"
    )
    cls.ff_cuda_dump_url = DataUrl(
        file_path=path.join(cls.dname, "dff_cuda.h5"), data_path="/entry/double_flatfield/results/data"
    )
    golden = 0
    for i in range(10):
        golden += exp(-i)
    cls.golden = golden / 10
    cls.tol = 1e-4

    if __has_pycuda__:
        cls.ctx = get_cuda_context(cleanup_at_exit=False)

    yield

    if __has_pycuda__:
        cls.ctx.pop()


@pytest.mark.usefixtures("bootstrap")
class TestDoubleFlatField:
    def test_dff_numpy(self):
        dff = DoubleFlatField(self.radios.shape, result_url=self.ff_dump_url)
        mydf = dff.get_double_flatfield(radios=self.radios)

        assert path.isfile(dff.result_url.file_path())

        dff2 = DoubleFlatField(self.radios.shape, result_url=self.ff_dump_url)
        mydf2 = dff2.get_double_flatfield(radios=self.radios)

        assert np.max(np.abs(mydf2 - mydf)) < self.tol
        assert np.max(np.abs(mydf - self.golden)) < self.tol

    @pytest.mark.skipif(not (__has_pycuda__), reason="Need pycuda for double flatfield with cuda backend")
    def test_dff_cuda(self):
        dff = CudaDoubleFlatField(self.radios.shape, result_url=self.ff_cuda_dump_url, cuda_options={"ctx": self.ctx})
        d_radios = garray.to_gpu(self.radios)
        mydf = dff.get_double_flatfield(radios=d_radios).get()

        assert path.isfile(dff.result_url.file_path())

        dff2 = CudaDoubleFlatField(self.radios.shape, result_url=self.ff_cuda_dump_url, cuda_options={"ctx": self.ctx})
        mydf2 = dff2.get_double_flatfield(radios=d_radios).get()

        assert np.max(np.abs(mydf2 - mydf)) < self.tol
        assert np.max(np.abs(mydf - self.golden)) < self.tol



import os
import posixpath
from multiprocessing.pool import ThreadPool
import numpy as np
from silx.io.dictdump import dicttonx, nxtodict
from ..misc.binning import binning as image_binning
from ..io.utils import get_first_hdf5_entry
from ..pipeline.config_validators import optional_tuple_of_floats_validator, optional_positive_integer_validator
from .cli_configs import ShrinkConfig
from .utils import parse_params_values


def access_nested_dict(dict_, path, default=None):
    items = [s for s in path.split(posixpath.sep) if len(s) > 0]
    if len(items) == 1:
        return dict_.get(items[0], default)
    if items[0] not in dict_:
        return default
    return access_nested_dict(dict_[items[0]], posixpath.sep.join(items[1:]))


def set_nested_dict_value(dict_, path, val):
    dirname, basename = posixpath.split(path)
    sub_dict = access_nested_dict(dict_, dirname)
    sub_dict[basename] = val


def shrink_dataset(input_file, output_file, binning=None, subsampling=None, entry=None, n_threads=1):
    entry = entry or get_first_hdf5_entry(input_file)
    data_dict = nxtodict(input_file, path=entry, dereference_links=False)

    to_subsample = [
        "control/data",
        "instrument/detector/count_time",
        "instrument/detector/data",
        "instrument/detector/image_key",
        "instrument/detector/image_key_control",
        "sample/rotation_angle",
        "sample/x_translation",
        "sample/y_translation",
        "sample/z_translation",
    ]

    detector_data = access_nested_dict(data_dict, "instrument/detector/data")
    if detector_data is None:
        raise ValueError("No data found in %s entry %s" % (input_file, entry))

    if binning is not None:

        def _apply_binning(img_res_tuple):
            img, res = img_res_tuple
            res[:] = image_binning(img, binning)

        data_binned = np.zeros(
            (detector_data.shape[0], detector_data.shape[1] // binning[0], detector_data.shape[2] // binning[1]),
            detector_data.dtype,
        )
        with ThreadPool(n_threads) as tp:
            tp.map(_apply_binning, zip(detector_data, data_binned))
        detector_data = data_binned
        set_nested_dict_value(data_dict, "instrument/detector/data", data_binned)

    if subsampling is not None:
        for item_path in to_subsample:
            item_val = access_nested_dict(data_dict, item_path)
            if item_val is not None:
                set_nested_dict_value(data_dict, item_path, item_val[::subsampling])
    dicttonx(data_dict, output_file, h5path=entry)


def shrink_cli():
    args = parse_params_values(ShrinkConfig, parser_description="Shrink a NX dataset")

    if not (os.path.isfile(args["input_file"])):
        print("No such file: %s" % args["input_file"])
        exit(1)
    if os.path.isfile(args["output_file"]):
        print("Output file %s already exists, not overwriting it" % args["output_file"])
        exit(1)

    binning = optional_tuple_of_floats_validator("", "binning", args["binning"])  # pylint: disable=E1121
    if binning is not None:
        binning = tuple(map(int, binning))
    subsampling = optional_positive_integer_validator("", "subsampling", args["subsampling"])  # pylint: disable=E1121

    shrink_dataset(
        args["input_file"],
        args["output_file"],
        binning=binning,
        subsampling=subsampling,
        entry=args["entry"],
        n_threads=args["threads"],
    )
    return 0


if __name__ == "__main__":
    shrink_cli()



import numpy as np


def is_fullturn_scan(angles_rad, tol=None):
    """
    Return True if the angles correspond to a full-turn (360 degrees) scan.
    """
    angles_rad = np.sort(angles_rad)
    if tol is None:
        tol = np.min(np.abs(np.diff(angles_rad))) * 1.1
    return np.abs((angles_rad.max() - angles_rad.min()) - (2 * np.pi)) < tol
